from .tensordict import TensorDict, stack, lag_tensor
from .testing import assert_tensors_allclose
import torch
import pytest


def _get_tensordict_example():
    n = 10
    shape = (1, n)
    return TensorDict({'a': torch.ones(*shape)})


def _get_tensordict_from_shapes(shape1, shape2):
    return TensorDict({'a': torch.rand(shape1), 'c': torch.rand(shape2)})


@pytest.mark.parametrize('shape1, shape2, dim, raises_error', [
    ([4], [4], 0, False),
    ([3], [4], 0, True),
    ([1, 3], [1, 3], 0, False),
    ([1, 3], [1, 4], 0, False),
    ([1, 3], [1, 3], 1, False),
])
def test_tensordict_size(shape1, shape2, dim, raises_error):
    d = _get_tensordict_from_shapes(shape1, shape2)
    if raises_error:
        with pytest.raises(ValueError):
            d.size(dim)
    else:
        assert d.size(dim) == shape1[dim]


def test_tensordict():
    b = {'a': torch.rand(1), 'c': torch.rand(1)}
    a = TensorDict(b)
    assert b.keys() == a.keys()

    # test dict
    for key in b:
        assert a[key] == b[key]

    # test arithmetic operations
    s = a + a
    assert isinstance(s, TensorDict)
    assert s['a'].item() == (a['a'] + a['a']).item()

    s = a / a
    assert isinstance(s, TensorDict)
    assert s['a'].item() == (a['a'] / a['a']).item()

    s = a * a
    assert s['a'].item() == (a['a'] * a['a']).item()

    assert s.keys() == a.keys()

    # test scalar ops
    assert isinstance(2 * a, TensorDict)
    assert (2 * a)['a'].item() == 2 * a['a'].item()
    assert (a * 2)['a'].item() == 2 * a['a'].item()


def test_tensordict_needs_compatible_keys():
    a = TensorDict({'a': 1, 'c': 1})
    b = TensorDict({'a': 1})
    with pytest.raises(ValueError):
        a + b


def test_tensordict_copy():
    a = TensorDict({})
    assert isinstance(a.copy(), TensorDict)


def test_tensordict_apply():
    a = TensorDict({'a': 1})
    b = a.apply(lambda x: 2 * x)
    assert isinstance(b, TensorDict)
    assert b['a'] == 2


def test_tensordict_shape():
    n = 10
    shape = (1, n)
    a = TensorDict({'a': torch.ones(*shape)})
    assert a['a'].shape == shape


def test_tensordict_split():
    n = 10
    shape = (1, n)
    a = TensorDict({'a': torch.ones(*shape)})

    splits = a.split(1, dim=1)

    for split in splits:
        assert isinstance(split, TensorDict)
        # test that key is present
        split['a']
        assert split['a'].shape == (1, 1)


tensor = _get_tensordict_example()['a']
@pytest.mark.parametrize('attr', dir(tensor))
def test_tensordict_dispatch(attr):
    if attr in ['imag', 'real']:
        pytest.xfail()
    t = _get_tensordict_example()
    a = t['a']
    getattr(t, attr)


def test_stack():
    t = _get_tensordict_example()
    n_stack = 2
    shape = t['a'].shape
    out = stack([t] * n_stack, dim=0)
    assert isinstance(out, TensorDict)
    out['a']
    assert out['a'].shape == (n_stack,) + shape


def test_lag_tensor():
    a = torch.arange(10)
    b = lag_tensor(a, 1, 0)
    assert_tensors_allclose(a[1:], b)

    b = lag_tensor(a, -1, 0)
    assert_tensors_allclose(a[:-1], b)



"""Integration tests"""
import numpy as np
import pytest

import xarray as xr
from torch.utils.data import DataLoader
from uwnet.ml_models.nn.datasets_handler import XRTimeSeries
from uwnet.utils import batch_to_model_inputs


def _mock_dataset(dtype=np.float32):
    prognostic = [['b', 5]]
    auxiliary = [['a', 1]]
    diagnostic = [['c', 5]]
    forcing = ['b']

    inputs = prognostic + auxiliary
    outputs = prognostic + diagnostic

    loss_scale = {'b': 1.0, 'c': 1.0}

    t, z, y, x = (10, 5, 4, 3)

    dims_full = ['time', 'z', 'y', 'x']
    dims_2d = ['time', 'y', 'x']

    # make dataset
    data_vars = {}
    for name, num in inputs + outputs:
        if num == 1:
            data_vars[name] = (dims_2d, np.zeros((t, y, x), dtype=dtype))
        else:
            data_vars[name] = (dims_full, np.zeros(
                (t, num, y, x), dtype=dtype))

    coords = {
        dim: np.arange(shape, dtype=dtype)
        for dim, shape in zip(dims_full, (t, z, y, x))
    }

    data_vars['layer_mass'] = ('z', np.ones((z, ), dtype=dtype))
    data_vars['Fb'] = data_vars['b']

    ds = xr.Dataset(data_vars, coords=coords)
    ds.time.attrs['units'] = 's'

    return prognostic, auxiliary, diagnostic, forcing, loss_scale, ds


@pytest.mark.skip()
def test_datasets_model_integration():
    #TODO use the new model for this integration test
    batch_size = 4

    # load the data
    prognostic, auxiliary, diagnostic, forcing, loss_scale, ds = _mock_dataset(
    )
    train_data = XRTimeSeries(ds.load())
    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    constants = train_data.torch_constants()

    # initialize model
    lstm = ForcedStepper(
        train_data.mean,
        train_data.scale,
        train_data.timestep(),
        auxiliary=auxiliary,
        prognostic=prognostic,
        diagnostic=diagnostic,
        forcing=forcing)

    # get one batch from the data loader
    x = next(iter(train_loader))
    x = batch_to_model_inputs(x, auxiliary, prognostic, diagnostic, forcing,
                              constants)

    # run the outputs through the network
    y = lstm(x, n=1)

    # compute the loss
    criterion = MVLoss(loss_scale.keys(), constants['layer_mass'], loss_scale)

    print("Loss:", criterion(y, x).item())



import torch
from torch import nn
from .modules import LinearDictIn, LinearDictOut, LinearFixed, MapByKey
from .tensordict import TensorDict
import pytest
from .testing import assert_tensors_allclose


def test_LinearDictIn():
    inputs = [('A', 10), ('B', 5)]
    n_out = 4

    lin = LinearDictIn(inputs, n_out)

    data = {'A': torch.zeros(1, 10), 'B': torch.zeros(1, 5)}
    out = lin(data)
    assert out.size(-1) == n_out
    out[0, 0].backward()


def test_LinearDictOut():
    outputs = [('A', 10), ('B', 5)]
    n_in = 4

    input = torch.zeros(1, n_in)
    lin = LinearDictOut(n_in, outputs)
    out = lin(input)

    for key, num in outputs:
        assert out[key].size(-1) == num

    out['A'][0, 0].backward()


@pytest.mark.parametrize('n, m', [(10, 5), (10, 10)])
def test_LinearFixed(n, m):
    def identity(x):
        return x[:, :m]

    mod = LinearFixed.from_affine(identity, n)
    a = torch.rand(1, n)
    # need to use the same dtype
    mod = mod.to(a.dtype)
    b = mod(a)
    assert b.shape == (1, m)
    return assert_tensors_allclose(identity(a), b)


def test_MapByKey():
    """Test that the module works"""
    name = 'a'
    n = 10

    class Mock(nn.Module):
        in_features = n

        def forward(self, x):
            return x

    funcs = {name: Mock()}
    data = {name: torch.rand(10)}
    mod = MapByKey(funcs)
    # test that forward evaluation works
    out = mod(data)
    out[name]  # out should be a dict
    assert isinstance(out, TensorDict)
    assert_tensors_allclose(out[name], data[name])

    # test inputs property
    assert mod.inputs == [(name, n)]



import logging
from functools import reduce

from toolz.curried import valmap
from toolz import first

import torch
from torch import nn

import xarray as xr

logger = logging.getLogger(__name__)


def _scale_var(scale, mean, x):
    x = x.double()
    mu = mean.double()
    sig = scale.double()

    x = x.sub(mu)
    x = x.div(sig + 1e-7)

    return x.float()


def _dict_to_parameter_dict(x):
    out = {}
    for key in x:
        out[key] = nn.Parameter(x[key], requires_grad=False)
    return nn.ParameterDict(out)


def _convert_dataset_to_torch_dict(ds: xr.Dataset):
    return {key: torch.from_numpy(ds[key].values)
            for key in ds.data_vars}


def _compute_scaler_data_from_xarray(dataset):
    logger.info("Computing mean")
    mean = dataset.mean(['x', 'y', 'time'])
    logger.info("Computing std")
    scale = dataset.std(['x', 'y', 'time'])

    args = [valmap(torch.squeeze, _convert_dataset_to_torch_dict(arg))
            for arg in [mean, scale]]
    return args


def add_tuples(x, y):
    return tuple(xx + yy for xx, yy in zip(x, y))


def moments(data):
    """Compute the first and second moments over first two dimensions

    Parameters
    ----------
    data : TensorDict

    Returns
    -------
    num_examples, ex, ex2
    """
    data = data.double()
    first_variable = first(data.values())
    shape = first_variable.shape
    num_examples = shape[1] * shape[0]
    m1 = data.sum(0).sum(0)
    m2 = (data ** 2).sum(0).sum(0)
    return num_examples, m1, m2


def moments_from_data_loader(loader):
    n, m1, m2 = reduce(add_tuples, map(moments, iter(loader)))
    m1 = m1 / n
    v = m2 / n - m1 ** 2
    return m1, v.sqrt()


class Scaler(nn.Module):
    """Torch class for normalizing data along the final dimension"""

    def __init__(self, mean=None, scale=None):
        "docstring"
        super(Scaler, self).__init__()
        self.mean = mean
        self.scale = scale

    def forward(self, x):
        out = {}
        for key in x:
            if key in self.scale and key in self.mean:
                out[key] = _scale_var(self.scale[key], self.mean[key], x[key])
            else:
                out[key] = x[key]
        return out

    def set_mean_scale(self, mean, scale):
        self.mean = _dict_to_parameter_dict(mean)
        self.scale = _dict_to_parameter_dict(scale)

    def fit_xarray(self, dataset):
        mean, scale = _compute_scaler_data_from_xarray(dataset)
        self.set_mean_scale(mean, scale)
        return self

    def fit_generator(self, data_loader):
        mean, sig = moments_from_data_loader(data_loader)
        self.set_mean_scale(mean.view(-1), sig.view(-1))
        return self



import torch
from torch.autograd import grad
from .xarray_interface import dataset_to_torch_dict
import numpy as np


def jacobian_from_xarray(model, ds):
    torch_data = dataset_to_torch_dict(ds)
    torch_data = torch_data[model.input_names]

    # compute saliency map
    jac = jacobian_from_model(model, torch_data)

    return jac


def dict_format_to_numpy(jac):
    out = []
    for outkey in jac:
        row = []
        for inkey in jac:
            arr = jac[outkey][inkey].detach().numpy()
            row.append(arr)
        out.append(row)
    return np.block(out)


def jacobian_backward(y, x):
    """Back-propagates the Frobenious norm of the jacobian"""
    n = len(y)
    out = 0.0
    for i in range(n):
        y_x = grad(y[i], x, create_graph=True)[0]
        y_x2 = y_x.norm() ** 2 / 2
        y_x2.backward(retain_graph=True)
        out += y_x2.item()
    return out


def jacobian_norm(y, x):
    n = len(y)
    out = 0.0
    for i in range(n):
        y_x = grad(y[i], x, create_graph=True)[0]
        out += y_x.norm() ** 2 / 2
    return out


def jacobian(y, x):
    """Jacobian of y with respect to x

    Returns
    -------
    jacobian : torch.tensor
        jacobian[i,j] is the partial derivative of y[i] with respect to x[j].
    """
    n = len(y)
    jac = []
    for i in range(n):
        y_x = grad(y[i], x, create_graph=True, allow_unused=True)[0]
        if y_x is None:
            y_x = torch.zeros(x.size(0))
        jac.append(y_x)
    return torch.stack(jac)


def max_eig_val(A, niter=10, m=1):
    """

    Parameters
    ----------
    A : matrix
    niter : number of iterations of power method
    m :
        number of iterations to keep gradients from end to keep gradients for
    """
    n = A.size(0)
    x = torch.rand(n)
    for i in range(niter):
        if i < niter - m:
            x = x.detach()
        y = A.matmul(x)
        norm = x.norm()
        lam = y.dot(x) / norm / norm
        x = y / lam / norm
    return lam, x


def max_signed_eigvals(A, niter=100, m=1):

    # find maximum norm eigvalue
    lam, _ = max_eig_val(A, niter=niter, m=m)
    # if it is negative shift the matrix
    h = -1 / lam * 0.9

    I = torch.eye(A.size(0))
    B = I + h * A

    lam_plus, _ = max_eig_val(B, niter=niter, m=m)
    lam_orig = (lam_plus - 1) / h

    if lam.item() < lam_orig.item():
        lam, lam_orig = lam_orig, lam
    return lam, lam_orig


def dict_jacobian(y, d, progs=["QT", "SLI"]):
    """Compute Jacobian dictionary

    Returns
    -------
    jacobian : dict of dicts
        jacobian[outkey][inkey][i, j] is the sensitivity of
        outkey[i] with respect to inkey[j]
    """
    jac = {}
    for inkey in d:
        for outkey in y:
            try:
                jac.setdefault(outkey, {})[inkey] = jacobian(
                    y[outkey], d[inkey]
                ).squeeze()
            except KeyError:
                pass
    return jac


def jacobian_from_model(model, d, **kwargs):
    # enable gradients
    for key in d:
        d[key].requires_grad = True
    y = model(d)
    return dict_jacobian(y, d, **kwargs)



"""Debias a torch network using LASSO"""
import argparse

import numpy as np
import torch
import xarray as xr
from sklearn.linear_model import Lasso

from uwnet.thermo import compute_apparent_source
from uwnet.xarray_interface import XarrayWrapper
from uwnet.numpy_interface import NumpyWrapper


class LassoDebiasedModel(object):
    """Debias a model using LASSO

    Parameters
    ----------
    model : callable
    mapping : seq of tuples
        sequence of (prog_var, model_output_var, target_var) tuples. These
        are used to define the pointwise regression models.
    """

    def __init__(self, model, mapping):
        "docstring"
        self.model = model
        self.coefs_ = {}
        self.mapping = mapping

    @property
    def xmodel(self):
        return XarrayWrapper(self.model)

    @property
    def npmodel(self):
        return NumpyWrapper(self.model)

    def fit(self, input):
        """

        Parameters
        ----------
        input : xr.Dataset
            input variables
        """
        output = self.xmodel(input)

        for args in self.mapping:
            print(f"Fitting regression model for {args}")
            prog, pred, target = args
            data = get_regression_problem(input, output, prog, pred, target)
            # These coefficients have shape (3, nz, ny)
            # reshaping to (3, nz, ny, 1) makes the broadcastable
            coefs = get_lasso_coefficients(data)
            self.coefs_[(prog, pred)] = coefs
        return self

    def predict(self, input):
        """
        Parameters
        ----------
        input : xr.Datset
        """
        output = XarrayWrapper(self.model)(input)

        for (prog, pred), coefs in self.coefs_.items():
            debiased = (
                input[prog] * coefs[0] + coefs[1] * output[pred] + coefs[2])
            output[pred] = debiased
        return output

    @property
    def numpy_coefs_(self):
        """Coefficients of LASSO
        """
        out = {}
        required_dim_order = ['dim_0', 'z', 'y']
        for key, coefs in self.coefs_.items():
            coefs = coefs.transpose(*required_dim_order)
            out[key] = coefs.values[..., np.newaxis]
        return out

    def predict_with_numpy(self, input):
        output = self.npmodel(input)

        for (prog, pred), coefs in self.numpy_coefs_.items():
            debiased = (
                input[prog] * coefs[0] + coefs[1] * output[pred] + coefs[2])
            output[pred] = debiased
        return output

    def __call__(self, input):
        return self.predict_with_numpy(input)


def insert_apparent_sources(ds, prognostics):
    sec_to_day = 86400
    for prog in prognostics:
        src = 'F' + prog
        target_key = 'Q' + prog
        target = compute_apparent_source(ds[prog], ds[src] * sec_to_day)
        ds = ds.assign(**{target_key: target})

    return ds


def get_regression_problem(input, output, prog, pred, target, time_dim='time'):

    regression_data = xr.Dataset({
        'prog': input[prog],
        'target': input[target],
        'pred': output[prog]
    })

    return regression_data.dropna(time_dim)


def get_lasso_coefficients(regression_data):
    out_dims = ['y', 'z']

    # input vars
    input_vars = ['prog', 'pred']
    output_var = 'target'

    def fit_linear(dfjk):
        df = dfjk.to_dataframe().reset_index()
        x_train = df[input_vars]
        y_train = df[output_var]

        debiaser = Lasso(alpha=.0001, normalize=True).fit(x_train, y_train)
        params = np.append(debiaser.coef_, debiaser.intercept_)
        return xr.DataArray(params)

    groupby = regression_data.stack(g=out_dims).groupby('g')
    params = groupby.apply(fit_linear)
    params = params.unstack('g')
    return params



from .tensordict import TensorDict
from .timestepper import Batch, TimeStepper
from .testing import assert_tensors_allclose
import torch


def _get_batch():
    n = 10
    array = torch.arange(n).view(1, -1)
    name = 'a'

    data = {name: array}
    return Batch(data, prognostics=[name]), name, array


def test_Batch_get_prognostics_at_time():
    batch, name, array = _get_batch()
    t = 1
    out = batch.get_prognostics_at_time(t)

    assert out[name].item() == array[0, t].item()


def test_TimeStepper():
    n_batch = 1
    n = 10
    array = torch.arange(n).view(1, -1).double()
    name = 'a'
    forcing_name = 'F' + name

    data = Batch(TensorDict({name: array, forcing_name: array * 0}),
                 prognostics=[name])
    one = TensorDict({name: torch.ones(n_batch, dtype=array.dtype)})

    def _mock_source(x):
        """Step by one each time"""
        return one

    stepper = TimeStepper(_mock_source, time_step=86400)
    output = stepper(data)

    assert output[name].shape == (1, n)
    assert_tensors_allclose(output['a'][:, :-1], array[:, 1:])



"""Convert netCDF dataset to a memory efficient layout for machine learning

The original dataset should have the dimensions ('time', 'z', 'y', 'x'). This
script combines the y and x dimensions into a sample dimension with a specified
chunk size. This sample dimension is optionally shufffled, and then saved to a
zarr archive.

The saved archive is suitable to use with uwnet.train
"""
import xarray as xr
import numpy as np

# arguments
input_file = snakemake.input[0]
output_file = snakemake.output[0]
variables = snakemake.params.variables
shuffle = snakemake.params.shuffle
train_or_test = snakemake.wildcards.train_or_test

chunk_size = 2**10

# open data
ds = xr.open_dataset(input_file)


if train_or_test == "train":
    ds = ds.isel(x=slice(0, 64))
elif train_or_test == "test":
    ds = ds.isel(x=slice(64, None))
else:
    raise NotImplementedError("{test_or_train} is not \"train\" or \"test\"")

# perform basic validation
assert ds['SLI'].dims == ('time', 'z', 'y', 'x')

# stack data
variables = list(variables) # needs to be a list for xarray
stacked = (ds[variables]
           .stack(sample=['y', 'x'])
           .drop('sample'))

# add needed variables
stacked['layer_mass'] = ds.layer_mass.isel(time=0)

# shuffle samples
if shuffle:
    n = len(stacked.sample)
    indices = np.random.choice(n, n, replace=False)
    stacked = stacked.isel(sample=indices)

chunked = stacked.chunk({'sample': chunk_size})

# save to disk
chunked.to_zarr(output_file)



import torch
from .jacobian import jacobian_backward, jacobian_norm, jacobian
import numpy as np
import pytest


def test_jacobian_backward():
    """Test the jacobian backprop function for a linear system

    y = A x

    For a linear system the gradient of Frobenious norm of the jacobian should
    be exactly equal to the original matrix.

    del_A  ||y_x||^2/2 = del_A || A ||^2  / 2 = A I = A


    """
    a = torch.rand(10, 10, requires_grad=True)
    x = torch.rand(10, requires_grad=True)

    y = a.matmul(x)
    jacobian_backward(y, x)

    A = a.grad.numpy()
    np.testing.assert_allclose(A, a.detach().numpy())


def test_jacobian_norm():

    a = torch.rand(10, 10, requires_grad=True)
    x = torch.rand(10, requires_grad=True)

    y = a.matmul(x)
    out = jacobian_norm(y, x)
    expected = a.norm() ** 2 / 2
    assert out.item() == pytest.approx(expected.item())

    # test gradient (see test_jacobian_backward docstring)
    out.backward()
    actual = a.grad.numpy()
    expected = a.data.numpy()
    np.testing.assert_allclose(actual, expected)


def test_jacobian():

    a = torch.rand(10, 10, requires_grad=True)
    x = torch.rand(10, requires_grad=True)

    y = a.matmul(x)
    out = jacobian(y, x)
    actual = out.data.numpy()
    expected = a.data.numpy()
    np.testing.assert_allclose(actual, expected)



import functools
from abc import ABCMeta
from collections import KeysView, MutableMapping, defaultdict

import torch
import numpy as np
from toolz import curry, first

mathfuns = [
    '__add__', '__rmul__', '__mul__', '__sub__', '__radd__', '__rmul__',
    '__truediv__', '__floordiv__', '__pow__', '__rpow__'
]

dictfuns = ['__len__', '__iter__', '__delitem__', '__setitem__']


@curry
def valmap_binary_operator(name, self, other):
    out = {}
    for key in self.data:
        func = getattr(self.data[key], name)
        if isinstance(other, self.__class__):
            out[key] = func(other.data[key])
            if self.keys() != other.keys():
                raise ValueError("Both arguments to binary arguments must "
                                 "have the same keys.")
        else:
            out[key] = func(other)
    return TensorDict(out)


@curry
def valmap_unary_operator(name, attrname, self, *args, **kwargs):
    func = getattr(getattr(self, attrname), name)
    return func(*args, **kwargs)


class ArithmaticMeta(ABCMeta):
    def __new__(cls, name, bases, dct):
        for fun in mathfuns:
            dct[fun] = valmap_binary_operator(fun)
        for fun in dictfuns:
            dct[fun] = valmap_unary_operator(fun, 'data')
        return super().__new__(cls, name, bases, dct)


class TensorDict(MutableMapping, metaclass=ArithmaticMeta):
    """Wrapper which overloads operators for dicts of tensors"""

    def __init__(self, data):
        self.data = data

    def __getitem__(self, key):
        if isinstance(key, (list, set, KeysView, self.__class__)):
            keys = list(key)
            return TensorDict({key: self.data[key] for key in keys})
        else:
            return self.data[key]

    def __getattr__(self, key):
        """Get attribute from the values"""
        if 'data' in self.__dict__:
            object = first(self.data.values())
        else:
            raise AttributeError()

        @functools.wraps(getattr(object, key))
        def fun(*args, **kwargs):
            return self.apply(lambda x: getattr(x, key)(*args, **kwargs))

        return fun

    def copy(self):
        return TensorDict(self.data.copy())

    def apply(self, fun):
        out = self.copy()
        for key in self:
            out[key] = fun(self[key])
        return out

    def __repr__(self):
        s = "TensorDict:\n"
        s += "-----------\n"
        for key in self.keys():
            shape = tuple(self.data[key].shape)
            dtype = self.data[key].dtype
            s += f"{key} \t({dtype}): {shape}\n"
        return s

    @functools.wraps(torch.split)
    def split(self, *args, **kwargs):
        outputs = []
        vals = (val.split(*args, **kwargs) for val in self.data.values())
        for vals in zip(*vals):
            d = dict(zip(self.keys(), vals))
            outputs.append(TensorDict(d))
        return outputs

    def size(self, dim):
        """The size of a dimension if it is the same for each component tensor
        """
        answers = {val.size(dim) for key, val in self.items()}
        if len(answers) == 1:
            return first(answers)
        else:
            raise ValueError(
                f"the sizes are unique along dimension {dim}")

    @classmethod
    def from_numpy_dict(cls, data):
        data_dict = {key: torch.from_numpy(np.asarray(arr))
                     for key, arr in data.items()}
        return cls(data_dict)


def stack(seq, dim=0):
    """Stack tensordicts"""
    data = defaultdict(list)

    for item in seq:
        for key, val in item.items():
            data[key].append(val)

    return TensorDict({key: torch.stack(data[key], dim=dim) for key in data})


def lag_tensor(arr: torch.tensor, lag, dim):
    if lag == 0:
        return arr
    elif lag > 0:
        selector = slice(lag, None)
    elif lag < 0:
        selector = slice(0, lag)
    index = [slice(None)] * arr.dim()
    index[dim] = selector

    return arr[index]


def lag(td, lag, dim):
    """Lag the arrays in a tensordict"""
    return td.apply(lambda x: lag_tensor(x, lag, dim))




import fsspec
import pickle
import logging
import numpy as np


width = 6


def load_pickle_from_url(url):
    logging.info(f"Opening {url}")
    openfile = fsspec.open(url, mode="rb")
    with openfile as f:
        s = f.read()
    return pickle.loads(s)


def add_wavelength_colorbar(fig, im, **kwargs):

    wave_length = np.array([10, 100, 200, 300, 500, 1000, 10000])
    tick_locations = 2 * np.pi / wave_length / 1e3

    cb = fig.colorbar(im, ticks=tick_locations, **kwargs)
    cb.ax.set_yticklabels(wave_length)
    cb.set_label('Wavelength (km)')
    return cb



from dataclasses import dataclass

import numpy as np
import torch
from toolz import valmap
from torch.utils.data import Dataset
from itertools import product
import xarray as xr

from uwnet.utils import dataset_to_broadcastable_array_dict
from uwnet.tensordict import TensorDict
from src.data import assign_apparent_sources

from uwnet.thermo import sec_in_day

def _stack_or_rename(x, **kwargs):
    for key, val in kwargs.items():
        if isinstance(val, str):
            x = x.rename({val: key})
        else:
            x = x.stack(**{key: val})
    return x


def _ds_slice_to_numpy_dict(ds):
    out = {}
    for key in ds.data_vars:
        out[key] = _to_numpy(ds[key])
    return out


def _to_numpy(x: xr.DataArray):
    dim_order = ['xbatch', 'xtime', 'xfeat']
    dims = [dim for dim in dim_order if dim in x.dims]
    return x.transpose(*dims).values


def _numpy_to_torch(x):
    y = torch.from_numpy(x).detach().float()
    return y.view(-1, 1, 1)


def _ds_slice_to_torch(ds):
    return valmap(_numpy_to_torch, _ds_slice_to_numpy_dict(ds))


class XRTimeSeries(Dataset):
    """A pytorch Dataset class for time series data in xarray format

    This function assumes the data has dimensions ['time', 'z', 'y', 'x'], and
    that the axes of the data arrays are all stored in that order.

    An individual "sample" is the full time time series from a single
    horizontal location. The time-varying variables in this sample will have
    shape (time, z, 1, 1).

    Examples
    --------
    >>> ds = xr.open_dataset("in.nc")
    >>> dataset = XRTimeSeries(ds)
    >>> dataset[0]

    """
    dims = ['time', 'z', 'x', 'y']

    def __init__(self, data, time_length=None):
        """
        Parameters
        ----------
        data : xr.DataArray
            An input dataset. This dataset must contain at least some variables
            with all of the dimensions ['time' , 'z', 'x', 'y'].
        time_length : int, optional
            The length of the time sequences to use, must evenly divide the
            total number of time points.
        """
        self.time_length = time_length or len(data.time)
        self.data = data
        self.numpy_data = {key: data[key].values for key in data.data_vars}
        self.data_vars = set(data.data_vars) - {'p', 'RHO', 'rho', 'Ps', 'layer_mass'}
        self.dims = {key: data[key].dims for key in data.data_vars}
        self.constants = {
            key
            for key in data.data_vars
            if len({'x', 'y', 'time'} & set(data[key].dims)) == 0
        }
        self.setup_indices()

    def setup_indices(self):
        len_x = len(self.data['x'].values)
        len_y = len(self.data['y'].values)
        len_t = len(self.data['time'].values)

        x_iter = range(0, len_x, 1)
        y_iter = range(0, len_y, 1)
        t_iter = range(0, len_t, self.time_length)
        assert len_t % self.time_length == 0
        self.indices = list(product(t_iter, y_iter, x_iter))

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, i):
        t, y, x = self.indices[i]
        output_tensors = {}
        for key in self.data_vars:
            if key in self.constants:
                continue

            data_array = self.numpy_data[key]
            if 'z' in self.dims[key]:
                this_array_index = (slice(t, t + self.time_length),
                                    slice(None), y, x)
            else:
                this_array_index = (slice(t, t + self.time_length), None, y, x)

            sample = data_array[this_array_index][:, :, np.newaxis, np.newaxis]
            output_tensors[key] = sample.astype(np.float32)
        return output_tensors

    @property
    def time_dim(self):
        return self.dims[0][0]

    def torch_constants(self):
        return {
            key: torch.tensor(self.data[key].values, requires_grad=False)
            .float()
            for key in self.constants
        }

    @property
    def scale(self):
        std = self.std
        return valmap(lambda x: x.max(), std)



@dataclass
class XarrayBatchLoader:
    """Yield batches from an xarray"""
    dataset: xr.Dataset
    batch_size: int
    dims: tuple = ('sample', 'time', 'z', 'x', 'y')
    variables: list = None
    torch: bool = False

    @property
    def batches(self):
        n = self.num_samples
        batch_size = self.batch_size
        indices = list(range(0, n, batch_size)) + [n]
        batches = [slice(i_start, i_end)
                   for i_start, i_end in zip(indices[:-1], indices[1:])]
        return batches


    @property
    def num_samples(self):
        return len(self.dataset.sample)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        for batch in self.batches:
            if self.variables is None:
                variables = list(self.dataset.data_vars)
            else:
                variables = self.variables
            subset = self.dataset[variables].isel(sample=batch)
            numpy_dict = dataset_to_broadcastable_array_dict(subset, dims=self.dims)
            if self.torch:
                yield TensorDict.from_numpy_dict(numpy_dict)
            else:
                yield numpy_dict

def get_timestep(data):
    time_dim = 'time'
    time = data[time_dim]
    dt = np.diff(time)

    all_equal = dt.std() / dt.mean() < 1e-6
    if not all_equal:
        raise ValueError("Data must be uniformly sampled in time")

    if time.units.startswith('d'):
        return dt[0] * 86400
    elif time.units.startswith('s'):
        return dt[0]
    else:
        raise ValueError(
            f"Units of time are {time.units}, but must be either seconds"
            "or days")


def get_dataset(data, predict_radiation):
    # _log.info("Opening xarray dataset")

    try:
        dataset = xr.open_zarr(data)
    except ValueError:
        dataset = xr.open_dataset(data)

    if not predict_radiation:
        dataset['FSLI'] = dataset['FSLI'] + dataset['QRAD'] / sec_in_day

    try:
        return dataset.isel(step=0).drop('step').drop('p')
    except:
        return dataset


def get_data_loader(
    ds: xr.Dataset,
    prognostics, 
    batch_size
):
    # List needed variables
    variables = prognostics + ['SST', 'SOLIN', 'QRAD']
    for variable in prognostics:
        forcing_key = 'F' + variable
        variables.append(forcing_key)

    train_data = XarrayBatchLoader(ds, batch_size=batch_size, variables=variables, torch=True)
    return train_data


"""A module containing useful patches to xarray
"""
import functools
import inspect
from functools import reduce
from operator import mul

import dask.array as da
import numpy as np
import scipy.ndimage
import xarray as xr
from scipy.interpolate import interp1d


# ndimage wrapper
class MetaNdImage(type):
    def __new__(cls, name, parents, dct):

        # for each function in scipy.ndimage wrap and add to class
        for func_name, func in inspect.getmembers(scipy.ndimage, inspect.isfunction):
            if func_name[-2:] == '1d':
                dct[func_name] = MetaNdImage.wrapper1d(func)
            else:
                dct[func_name] = MetaNdImage.wrappernd(func)
            # setattr(xr.DataArray, 'ndimage_' + func_name, ndimage_wrapper(func))
        return super(MetaNdImage, cls).__new__(cls, name, parents, dct)


    def wrappernd(func):
        """Wrap a subset of scipy.ndimage functions for easy use with xarray"""

        @functools.wraps(func)
        def f(self, axes_kwargs, *args, dims=[], **kwargs):

            x = self._obj
            # named axes args to list
            axes_args = [axes_kwargs[k] for k in x.dims]
            y = x.copy()

            axes_args.extend(args)
            y.values = func(x, axes_args, **kwargs)
            y.attrs['edits'] = repr(func.__code__)

            return y

        return f


    def wrapper1d(func):
        """Wrapper for 1D functions
        """

        @functools.wraps(func)
        def f(self, dim, *args, **kwargs):

            x = self._obj
            # named axes args to list
            y = x.copy()
            y.values = func(x, *args, axis=x.get_axis_num(dim), **kwargs)
            y.attrs['edits'] = repr(func.__code__)

            return y

        return f


@xr.register_dataarray_accessor('ndimage')
class NdImageAccesor(metaclass=MetaNdImage):
    def __init__(self, obj):
        self._obj = obj



import fsspec
import xarray as xr
from .thermo import compute_apparent_source

TRAINING_DATA_PATH = "gs://vcm-ml-data/project_data/uwnet/2020-02-01-noBlur.zarr/"


def open_data(tag):
    if tag == 'training':
        return xr.open_dataset("../../data/processed/training/noBlur.nc")
        return xr.open_zarr(fsspec.get_mapper(TRAINING_DATA_PATH))


def assign_apparent_sources(ds):
    return ds.assign(
        Q1=compute_apparent_source(ds.SLI, 86400 * ds.FSLI),
        Q2=compute_apparent_source(ds.QT, 86400 * ds.FQT))



import numpy as np
import torch
from uwnet.utils import stack_dicts, dataarray_to_broadcastable_array
import xarray as xr


def test_stack_dicts():

    n = 10
    ins = [{'a': torch.ones(2, 1)} for i in range(n)]

    out = stack_dicts(ins)
    assert out['a'].size() == (2, n, 1)


def test_dataarray_to_broadcastable_array():

    a, b = 4, 5
    dims = ['a', 'b']
    arr = xr.DataArray(np.ones((a, b)), dims=dims)
    desired_dims = ['c', 'a', 'b']
    expected = (1, a, b)
    nparr = dataarray_to_broadcastable_array(arr, desired_dims)
    assert nparr.shape == expected


    a, b = 4, 5
    dims = ['b', 'a']
    arr = xr.DataArray(np.ones((b, a)), dims=dims)
    expected = (1, a, b)
    nparr = dataarray_to_broadcastable_array(arr, desired_dims)
    assert nparr.shape == expected



"""

Step Types:

- instability
- multi

"""
import torch
from toolz import curry
from torch.nn.functional import mse_loss

from uwnet.utils import mean_other_dims
from .timestepper import predict_multiple_steps, TimeStepper, Batch
from . import tensordict


def r2_score(truth, prediction):
    return weighted_r2_score(truth, prediction, 1.0, dim=None)


def weighted_r2_score(truth, prediction, weights, dim=-1):
    """Compute the weighted R2 score"""
    mean = mean_other_dims(truth, dim)
    squares = weighted_mean_squared_error(truth, mean, weights, dim)
    residuals = weighted_mean_squared_error(truth, prediction, weights, dim)
    return 1 - residuals / squares


@curry
def mse_with_integral(truth, prediction, weights, dim=-3):
    ctruth = (truth * weights).sum(dim) / 1000
    cpred = (prediction * weights).sum(dim) / 1000

    l1 = mse_loss(ctruth, cpred)
    l2 = weighted_mean_squared_error(truth, prediction, weights, dim)

    return l1 + l2


@curry
def weighted_mean_squared_error(truth, prediction, weights=1.0, dim=-1):
    """Compute the weighted mean squared error

    Parameters
    ----------
    truth: torch.tensor
    prediction: torch.tensor
    weights: torch.tensor
        one dimensional tensor of weights. Must be the same length as the truth
        and prediction arrays along the dimension `dim`
    dim:
        the dimension the data should be weighted along
    """
    error = truth - prediction
    error2 = error * error
    error2 = mean_other_dims(error2, dim)
    return torch.mean(error2 * weights)


def select_keys_time(x, keys, t):
    return {key: x[key][t] for key in keys}


def compute_loss(criterion, prognostics, y):
    return sum(criterion(prognostics[key], y[key]) for key in prognostics)


def compute_multiple_step_loss(criterion, model, batch, *args, **kwargs):
    """Compute the loss across multiple time steps with an Euler stepper

    Yields
    ------
    t: int
       the time step of the prediction
    prediction: dict
       the predicted state

    """
    prediction_generator = predict_multiple_steps(model, batch, *args,
                                                  **kwargs)
    return sum(
        compute_loss(criterion, prediction, batch.get_prognostics_at_time(t))
        for t, prediction, _ in prediction_generator)


def equilibrium_penalty(criterion, model, batch, dt, n=20):
    from random import randint
    i = randint(0, batch.num_time - 1)

    state0 = batch.get_prognostics_at_time(i)
    mean = batch.get_prognostics().apply(lambda x: x.mean(dim=1))
    g = batch.get_known_forcings()
    mean_forcing = g.apply(lambda x: x.mean(dim=1))
    state = state0

    for t in range(n):
        inputs = batch.get_model_inputs(i, state)
        src = model(inputs)
        state = state + dt * src + dt * mean_forcing * 86400

    return compute_loss(criterion, mean, state)


def instability_penalty_step(self, engine, data, alpha=1.0):
    # TODO pass the list of prognostics to step
    batch = Batch(data.float(), prognostics=['QT', 'SLI'])
    self.optimizer.zero_grad()

    dt = .125
    pred, x1 = get_input_output(self.model, dt, batch)

    l1 = compute_loss(self.criterion, x1, pred)
    l2 = equilibrium_penalty(self.criterion, self.model, batch, dt)

    loss = l1 + alpha * l2

    info = {
        'Q1/Q2': l1.item(),
        'equilibrium': l2.item(),
        'total': loss.item(),
    }

    loss.backward()
    self.optimizer.step()
    self.optimizer.zero_grad()
    engine.state.loss_info = info
    return pred, x1


def multiple_step_loss_step(self, engine, batch):
    self.optimizer.zero_grad()
    stepper = TimeStepper(self.model, time_step=self.time_step)

    prediction = stepper(batch)

    # TODO: maybe TimeStepper should return a batch object
    prediction = tensordict.lag(prediction, -1, batch.time_dim)
    truth = batch.data_for_lag(1)

    loss = ((truth - prediction)**2)
    combined_var_loss = sum(loss.values())
    loss = combined_var_loss[combined_var_loss < 100.0].mean()

    loss.backward()
    self.optimizer.step()
    self.optimizer.zero_grad()

    info = {"loss": loss.item()}

    engine.state.loss_info = info
    return info


@curry
def get_input_output(model, dt, batch):
    src = model(batch.data)

    g = batch.get_known_forcings()
    progs = batch.get_prognostics()

    forcing = g.apply(lambda x: (x[:, 1:] + x[:, :-1]) / 2)
    x0 = progs.apply(lambda x: x[:, :-1])
    x1 = progs.apply(lambda x: x[:, 1:])
    src = src.apply(lambda x: x[:, :-1])

    # pred = x0 + dt * src + dt * 86400 * forcing
    # return pred, x1

    # pred = x0 + dt * src + dt * 86400 * forcing
    # residual forcing
    residual = (x1-x0)/dt - 86400 * forcing
    return src, residual


def get_step(name, self, kwargs):
    if name == 'instability':
        fun = instability_penalty_step
    elif name == 'multi':
        fun = multiple_step_loss_step
    else:
        raise NotImplementedError(
            f"Training step type '{name}' is not implemented")

    return curry(fun, self, **kwargs)



"""Sample datasets
"""
import numpy as np
import xarray as xr


def tiltwave():
    """Periodic tilted wave"""

    x, y = np.ogrid[0:2*np.pi:201j, 0:np.pi:30j]
    z = np.sin(y)*np.cos(x) + np.cos(x-.5)*np.sin(2*y) + 2


    return xr.DataArray(z.T, dims=['z', 'x'], coords=dict(x=x.ravel(), z=y.ravel()))


def rand3d():
    """Random uniform 3d data"""
    sh = (100, 200, 300)
    dims = ['x', 'y', 'z']

    coords = {}
    for d, n in zip(dims, sh):
        coords[d] = np.arange(n)

    return xr.DataArray(np.random.rand(*sh), dims=dims, coords=coords)



"""Script for validating input data"""
import sys
import xarray as xr
import numpy as np
import click


def _check_nan(x):
    sum_nan = int(np.isnan(x).sum())
    if sum_nan != 0:
        raise ValueError(
            f"NaNs detected in input. Total number is {sum_nan} of {x.size}.")


def check_necessary_variables_present(ds):
    variables_needed = {
        'SOLIN', 'QT', 'SLI', 'FSLI', 'FQT', 'U', 'V', 'W', 'SOLIN', 'SST',
        'RADTOA', 'RADSFC', 'Prec', 'LHF', 'SHF'
    }

    not_present = variables_needed - set(ds.data_vars)
    if len(not_present) > 0:
        raise ValueError("Not all needed variables are present. These "
                         f"variables are missing: {not_present}.")


def check_for_nans(ds):
    for key in ds:
        _check_nan(ds[key])


def check_units(ds):
    for key, expected in [('FQT', 'g/kg/s'), ('FSLI', 'K/s')]:
        actual = ds[key].units
        if actual != expected:
            raise ValueError(f"{key} units are {actual}")


def check_w_correlated_with_fqt(ds):

    corr_spatial = (ds.W * ds.FQT).mean(['x', 'y', 'time'])
    ans = (corr_spatial * ds.layer_mass).sum('z')
    if ans < 0:
        raise ValueError(
            "FQT and W are negatively correlated...check this data.")


def check_time_equally_spaced(ds):
    time = ds.time
    dt = np.diff(time)

    sig = dt.std() / dt.mean()
    print(f"Variance in dt is {sig}%")
    assert sig < 1e-6


def check_w_domain_mean_vanishes(ds):
    # vertical velocity should vanish
    mean = ds.W[0].mean(['x', 'y'])
    sig = ds.W[0].std(['x', 'y'])

    M = ds.layer_mass.sum()

    mean_ss = np.sqrt(float((ds.layer_mass * (mean**2)).sum() / M))
    sig_ss = np.sqrt(float((ds.layer_mass * (sig**2)).sum() / M))

    if (mean_ss / sig_ss) > 1e-4:
        raise ValueError(
            "The domain mean of W is larger than acceptable. The typical "
            f"magnitude is {mean_ss} while the standard deviation is {sig_ss}."
        )


def run_checks(ds):
    check_funs = [f for f in globals() if f.startswith('check')]
    failed = 0
    for func_name in check_funs:
        func = globals()[func_name]
        print(f"Running {func_name}...")
        try:
            func(ds)
        except Exception as exc:
            click.secho("Check BAD", fg="red")
            print(exc)
            failed += 1
        else:
            click.secho("Check OK", fg="blue")

    if failed > 0:
        click.secho(f"Failed {failed} of {len(check_funs)} checks!", fg="red")
        sys.exit(1)
    else:
        click.echo("All tests passed!")


data = sys.argv[1]
try:
    ds = xr.open_zarr(data)
except ValueError:
    ds = xr.open_dataset(data)

run_checks(ds)



import numpy as np
import pytest
from functools import partial
from unittest.mock import Mock

from .wave import (
    WaveCoupler,
    LinearResponseFunction,
    _fill_zero_above_input_level,
    filter_small_eigenvalues,
    get_elliptic_matrix,
    WaveEq,
)


def mad(x, y):
    return np.abs(x - y).mean()


def order_of_convergence(n, err):
    return -np.polyfit(np.log(n), np.log(err), 2)[1]


def ans(m, H, c=2, n=32):
    """Method of manufactured solutions test"""
    # setup grid
    d = H / n
    z = np.arange(n) * d + d / 2
    zi = np.arange(n + 1) * d

    # setup solution
    rho = np.exp(-c * z / H)
    rhoi = np.exp(-c * zi / H)
    #     rhoi = centered_to_interface(rho)
    w = np.sin(m * np.pi * z / H)

    # setup rhs
    rhs = -c * np.pi * m / H ** 2 * np.cos(
        np.pi * m * z / H
    ) - np.pi ** 2 * m ** 2 / H ** 2 * np.sin(np.pi * m * z / H)

    # get matrix
    A = get_elliptic_matrix(rhoi, zi, rho, z)

    return A @ w, rhs


@pytest.mark.parametrize('test_soln', [
    partial(ans, m=3, H=1.0, c=5),
    partial(ans, m=3, H=1.0, c=0),
    partial(ans, m=1, H=16e3, c=0),
])
def test_elliptic_order_of_convergence(test_soln):

    err = []
    ns = [5, 10, 40, 100, 1000]
    for n in ns:
        pred, truth = test_soln(n=n)
        err.append(mad(pred, truth))
    assert order_of_convergence(ns, err) == pytest.approx(2.0, rel=0.1)


def test_numpy_linalg():
    """np.linalg.eig sometimes gives incorrect results

    This occured when using a docker image on the lambda box.
    """
    n = 10
    D = np.diag(np.arange(n))
    r, v = np.linalg.eig(D)
    I = np.eye(n)
    np.testing.assert_allclose(I, v)


def test_filter_small_eigenvalues():
    n = 10
    threshold = 4.5
    D = np.diag(np.arange(n))
    expected = np.where(D > threshold, D, 0.0)
    filtered = filter_small_eigenvalues(D, threshold)
    np.testing.assert_allclose(expected, filtered)


def create_jacobian(variables):
    # construct test data
    n = 5
    jacobian = {}
    for outkey in variables:
        jacobian[outkey] = {}
        for inkey in variables:
            jacobian[outkey][inkey] = np.ones((n, n))

    return jacobian


@pytest.mark.parametrize(
    "boundaries, variables",
    [({"a": 2, "b": 3}, ["a", "b"]), ({"a": 2, "b": 3}, ["a", "b", "c"])],
)
def test__fill_zero_above_input_level(boundaries, variables):
    jacobian = create_jacobian(variables)

    # call function
    ans = _fill_zero_above_input_level(jacobian, boundaries)

    # check answer
    for outkey in jacobian:
        for inkey in jacobian[outkey]:
            orig = jacobian[outkey][inkey]
            arr = ans[outkey][inkey]
            lid = boundaries.get(inkey, orig.shape[1])
            np.testing.assert_allclose(arr[:, lid:], 0.0)
            np.testing.assert_allclose(arr[:, :lid], orig[:, :lid])


def test_LinearResponseFunction_dump():
    from io import StringIO

    n = 10
    panes = {"s": {"s": np.random.rand(n, n)}}
    base = {"s": np.ones((n,))}
    lrf = LinearResponseFunction(panes, base)

    buffer = StringIO()
    lrf.dump(buffer)

    buffer.seek(0)
    lrf_loaded = LinearResponseFunction.load(buffer)

    for pane in lrf_loaded.iterpanes():
        assert isinstance(pane, np.ndarray)

    np.testing.assert_array_equal(lrf_loaded.panes["s"]["s"], lrf.panes["s"]["s"])


@pytest.fixture(params=[1e3, 5e3, 16e3])
def _test_wave(request):
    r"""A test wave problem with :math:`s = s_0 + cz` and :math:`\rho_0 = 1.0`


    Notes
    -----

    Let's test the results of this code on a simpler example with constant density, and
    :math:`s= s_0 + c z`.


    .. math::

        s_t + c w = 0\\
        u_t + p_x = 0 \\
        p_z = g \frac{s}{s_0}

    With some typical manipulations this can be written as one fourth order
    differential equation for w as follows:

    .. math::

        (\partial z^2 \partial t^2 + k^2 \frac{c g}{s_0}) w = 0

    The solution of this can be derived using separation of variables. Let
    :math:`w=Z(z)A(t)`, then

    .. math::

        Z'' A'' = k^2 \frac{c g}{s_0} Z A

    so that :math:`Z = sin(\pi z /H)`. for this mode:

    .. math::

        (\partial_t)^2 A + \lambda^2 A = 0 \\
        \lambda^2 = \frac{H^2 k^2 c g}{\pi^2 s_0}

    this is a wave equation with speed :math:`\lambda = \frac{H}{\pi} \sqrt{\frac{c}{s_0}}`
    """
    H = request.param

    n = 20
    d = H / n
    zc = np.arange(n) * d + d / 2
    zint = np.arange(n + 1) * d

    c = 0.005
    s0 = 300
    gravity = 9.81

    expected_speed = H / np.pi * np.sqrt(gravity * c / s0)

    density = np.ones(n)

    base_state = {
        "SLI": zc * c + s0,
        "QT": np.ones(n),
        "density": density,
        "height_interface": zint,
        "height_center": zc,
    }

    base_state = {key: val.astype(float) for key, val in base_state.items()}

    # need to patch in a constant in height buoyancy for this idealized test case to work
    def mock_buoyancy(s, q):
        return 9.81 * s / 300

    wave = WaveEq(base_state)
    wave.buoyancy = mock_buoyancy

    return wave, expected_speed


def test_wave_eq_correct_speed_integration(_test_wave):
    # This test is too complicated. Need to create a mock WaveEq object
    wave, expected_speed = _test_wave
    mock_lrf = Mock()
    mock_lrf.to_array.return_value = 0
    coupler = WaveCoupler(wave, mock_lrf)


    k = 3.0
    val = np.linalg.eigvals(coupler.system_matrix(k))
    growth_rate, phase_speed = val.real, val.imag / k

    assert growth_rate == pytest.approx(0.0)
    assert phase_speed.max() == pytest.approx(expected_speed, rel=.002)



import torch
from torch import nn
from toolz import first, pipe


def bucketize(tensor, bucket_boundaries):
    """Equivalent to numpy.digitize

    Notes
    -----
    Torch does not have a built in equivalent yet. I found this snippet here:
    https://github.com/pytorch/pytorch/issues/7284
    """
    result = torch.zeros_like(tensor, dtype=torch.int32)
    for boundary in bucket_boundaries:
        result += (tensor > boundary).int()
    return result


def _unflatten(output, shape):
    output_shape = shape + output.shape[1:]
    return output.view(*output_shape)


class GroupBy(object):
    def __init__(self, memberships):
        self.shape = memberships.shape
        self.memberships = memberships.view(-1)
        self.compute_group_indices()

    @property
    def n(self):
        return self.memberships.numel()

    @property
    def groups(self):
        return self.group_indices.keys()

    def compute_group_indices(self):
        num_indices = 0
        self.group_indices = {}
        k = 0
        while num_indices < self.n:
            inds = torch.nonzero(self.memberships == k).view(-1)
            self.group_indices[k] = inds
            num_indices += inds.shape[0]
            k += 1

    def _apply_over_groups(self, fun, x):
        outputs = {}
        for k, inds in self.group_indices.items():
            args = (k, x.index_select(0, inds))
            outputs[k] = fun(args)
        return outputs

    def _init_output_array(self, outputs):
        nout = first(outputs.values()).shape[1:]
        dtype = first(outputs.values()).dtype
        return torch.zeros(self.n, *nout, dtype=dtype)

    def _gather_from_group_outputs(self, outputs):
        output = self._init_output_array(outputs)
        for k in outputs:
            output[self.group_indices[k]] = outputs[k]
        return output

    def _unflatten(self, output):
        return _unflatten(output, self.shape)

    def flatten_input(self, x):
        m = len(self.shape)
        if x.shape[:m] != self.shape:
            raise ValueError(
                "memberships and x must match along the first few "
                "dimensions")

        if x.dim() <= m:
            raise ValueError("x must have more dimensions than memberships")

        shape = x.shape
        x = x.view(self.n, *shape[m:])
        return x

    def apply(self, fun, x):
        """Apply function over groups and combine output

        Parameters
        ----------
        fun : callable
            Takes (i, x) as an argument, output shape must include the group
            dimension.
        x : torch.tensor
            Argument to pass to fun. The first dimensions must match the
            membership array.

        Returns
        -------
        merged_output : torch.tensor
        """
        x = self.flatten_input(x)
        group_output = self._apply_over_groups(fun, x)
        return pipe(group_output, self._gather_from_group_outputs,
                    self._unflatten)


class DispatchByVariable(nn.Module):
    """Dispatch

    """

    def __init__(self, bins, objs, variable, index):
        super(DispatchByVariable, self).__init__()
        self.bins = bins
        self.objs = objs
        self.variable = variable
        self.index = index

    def get_binning_variable(self, x):
        return x[self.variable][..., self.index]

    def get_bin_membership(self, x):
        y = self.get_binning_variable(x)
        return bucketize(y, self.bins)

    def forward(self, x):
        memberships = self.get_bin_membership(x)
        return GroupBy(memberships, x)



import torch
from torch import nn

import uwnet.modules as um

from .xarray_interface import XRCallMixin


class InnerModel(nn.Module, XRCallMixin):
    """Inner model which operates with height along the last dimension"""

    def __init__(self, pre, post):
        "docstring"
        super(InnerModel, self).__init__()

        n = 256
        transpose = um.ValMap(um.RPartial(torch.transpose, -3, -1))
        self.input_names = [name for name, _ in pre.inputs]
        self.model = nn.Sequential(
            transpose,
            pre,
            um.LinearDictIn([(name, num) for name, num in pre.outputs], n),
            nn.ReLU(),
            nn.Linear(n, n),
            nn.ReLU(),
            nn.Linear(n, n),
            nn.ReLU(),
            um.LinearDictOut(n, [(name, num) for name, num in post.inputs]),
            post,
            transpose
        )

    def forward(self, x):
        return self.model(x[self.input_names])


def get_model(pre, post, _config):
    """Create an MLP with scaled inputs and outputs
    """
    kind = _config['kind']

    if kind == 'inner_model':
        return InnerModel(pre, post).to(dtype=torch.float)



from matplotlib import pyplot as plt
from matplotlib.colors import Normalize


def plot(args):
    """Plot jacobian

    Parameters
    ----------
    args : tuple
        (jac, p) tuple
    """
    fig, axs = plt.subplots(2, 2, constrained_layout=True, figsize=(5, 4))
    jac, p = args

    k = 0

    abc = 'abcd'

    progs = list(jac)

    for i, outkey in enumerate(progs):
        for j, inkey in enumerate(progs):
            ax = axs[i, j]
            val = jac[outkey][inkey] * 86400
            # select the colorbar norm
            # vmax = get_vmax(val)
#             do_sym_log = vmax > 10

#             if do_sym_log:
#                 linthresh = 1.0
#                 norm = SymLogNorm(
#                     linthresh=linthresh, linscale=20, vmin=-vmax, vmax=vmax)
#             else:
#                 norm = Normalize(vmin=-vmax, vmax=vmax)
#             if inkey == 'QT':
#                 norm = Normalize(vmin=-1.5, vmax=1.5)
#             else:
#                 norm =  Normalize(vmin=-1.0, vmax=1.0)

            im = ax.pcolormesh(p, p, val, cmap='RdBu_r')
            # ax.contourf(
            #     p, p, val, levels=[-100,-10,10, 100] , extend='both', colors='none',
            #     hatches=['xxxx', '...', None, '...', 'xxxx'],)

            letter = abc[k]
            ax.set_title(f"{letter}) d{outkey}/dt from {inkey}", loc='left')
            ax.set_xlabel(f"Pressure ({inkey})")
            ax.set_ylabel(f"Pressure (d{outkey}/dt)")
            ax.invert_yaxis()
            ax.invert_xaxis()

            cb = plt.colorbar(im, ax=ax, pad=-.04)

#             if do_sym_log:
#                 locator = matplotlib.ticker.SymmetricalLogLocator(
#                     linthresh=linthresh, base=10)
#                 locator.set_params(numticks=5)
#                 cb.locator = locator
#                 cb.update_ticks()

            k += 1

    return axs


#!/usr/bin/env python
# coding: utf-8
import json
from collections import Mapping, defaultdict
from dataclasses import asdict, dataclass
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy import sparse
from toolz import curry
from toolz.curried import compose, first, get, valmap

from ..jacobian import dict_jacobian
from .tensordict import TensorDict
from ..thermo import Lc, cp, grav, interface_heights

from . import utils


def centered_difference(x, dim=-3):
    """Compute centered difference with first order difference at edges"""
    x = x.transpose(dim, 0)
    dx = torch.cat(
        [
            torch.unsqueeze(x[1] - x[0], 0), x[2:] - x[:-2],
            torch.unsqueeze(x[-1] - x[-2], 0)
        ],
        dim=0)
    return dx.transpose(dim, 0)


def get_test_solution(base_state):
    from collections import namedtuple

    s, q = base_state["SLI"].detach(), base_state["QT"].detach()
    s.requires_grad = True
    q.requires_grad = True
    w = torch.zeros(s.size(0), requires_grad=True)
    soln = namedtuple("Solution", ["w", "s", "q"])
    return soln(w=w, s=s, q=q)


def ones(n):
    """Convenience function"""
    return torch.ones(n, requires_grad=True)


def pad_linear(z):
    return np.r_[2 * z[0] - z[1], z, 2 * z[-1] - z[-2]]


def centered_to_interface(z):
    z = pad_linear(z)
    return (z[1:] + z[:-1]) / 2


def vertically_integrate(f, z_interface):
    dz = z_interface[1:] - z_interface[:-1]
    return torch.cumsum(f * dz, 0)


def deriv(f, z_center):
    return centered_difference(f, 0) / centered_difference(z_center, 0)


# Routines for computing elliptic operators
def get_elliptic_diagonals(rhoi, zi, rho, z):
    rho = pad_linear(rho)
    z = pad_linear(z)
    dz = np.diff(z)
    dzi = np.diff(zi)

    a = rho[:-2] / dz[:-1] / dzi / rhoi[:-1]
    b = -rho[1:-1] / dzi * (1 / dz[1:] / rhoi[1:] + 1 / dz[:-1] / rhoi[:-1])
    c = rho[2:] / dz[1:] / dzi / rhoi[1:]

    at = a[1:]
    bt = b.copy()
    bt[0] = b[0] - a[0] * (zi[0] - z[0]) / (z[1] - zi[0])
    bt[-1] = b[-1] - c[-1] * (z[-1] - zi[-1]) / (zi[-1] - z[-2])
    ct = c[:-1]

    return at, bt, ct


def get_elliptic_matrix(*args):
    diags = get_elliptic_diagonals(*args)
    A = np.asarray(sparse.diags(diags, [-1, 0, 1]).todense())
    return A


def get_elliptic_matrix_easy(rho, z):
    """Get elliptic matrix from cell centered rho and z"""
    rho = np.asarray(rho)
    z = np.asarray(z)

    rhoi = centered_to_interface(rho)
    zi = centered_to_interface(z)

    return get_elliptic_matrix(rhoi, zi, rho, z)


# Functions for manipulating jacobian dictionaries #########################
def dict_to_matrix(d, order):
    matrix = []
    for out_key in order:
        row = np.concatenate([d[out_key][in_key] for in_key in order], axis=-1)
        matrix.append(row)
    return np.concatenate(matrix, axis=0)


def get_first_pane(jac):
    for i in jac:
        for j in jac[i]:
            return jac[i][j]


def ensure_panes_present(jac, required_keys):
    """Ensure that all required keys are present in an input dictionary"""
    output = {}
    # fill in zeros
    template = np.zeros_like(get_first_pane(jac))

    for i in required_keys:
        output[i] = {}
        for j in required_keys:
            try:
                output[i][j] = jac[i][j]
            except KeyError:
                output[i][j] = template
    return output


def _fill_zero_above_input_level(jac, boundaries):
    output = {}
    for outkey in jac:
        output[outkey] = {}
        for inkey in jac[outkey]:
            arr = jac[outkey][inkey]
            upper_lid = boundaries.get(inkey, arr.shape[1])
            i, j = np.ogrid[[slice(n) for n in arr.shape]]
            output[outkey][inkey] = np.where(j < upper_lid, arr, 0.0)
    return output


def subslice_blocks(jac, ind):
    out = jac.copy()
    for a in jac:
        for b in jac[a]:
            out[a][b] = jac[a][b][ind, ind]
    return out


def filter_small_eigenvalues(matrix: np.array, threshold: float, replace: float = 0):
    s, v = np.linalg.eig(matrix)
    s_filt = np.where(np.abs(s) > threshold, s, replace)
    return (v * s_filt) @ np.linalg.inv(v)
    # u, s, vh = np.linalg.svd(matrix)
    # s_filt = np.where(s> threshold, s, replace)
    # return (u * s_filt) @ vh


class WaveEq:

    field_order = ("w", "s", "q")

    def __init__(self, base_state):
        """

        Parameters
        ----------
        base_state
            dict of numpy arrays. Must have the following keys:
                'SLI', 'QT', 'density', 'height_center'
        """
        self.base_state = base_state
        self._base_state = utils.numpy2torch(base_state)

    @property
    def density(self) -> torch.Tensor:
        return self._base_state["density"]

    @property
    def center_heights(self) -> torch.Tensor:
        return self._base_state["height_center"]

    def buoyancy(self, s, q):
        grav = 9.81
        s0 = self._base_state["SLI"]
        return grav * (s - s0) / s0

    @property
    def elliptic_operator(self):
        rho = self.density.detach().numpy()
        z = self.center_heights.detach().numpy()
        return get_elliptic_matrix_easy(rho, z)

    @property
    def inverse_elliptic_operator(self):
        return torch.tensor(np.linalg.inv(self.elliptic_operator)).float()

    def invert_buoyancy(self, s, q):
        b = self.buoyancy(s, q)
        return -torch.matmul(self.inverse_elliptic_operator, b)

    @property
    def sz(self):
        return deriv(self._base_state["SLI"], self.center_heights)

    @property
    def qz(self):
        return deriv(self._base_state["QT"], self.center_heights)

    def advection_s(self, w):
        return -self.sz * w

    def advection_q(self, w):
        return -self.qz * w

    def get_test_solution(self):
        return get_test_solution(self._base_state)

    def system_matrix(self, k):
        soln = self.get_test_solution()
        outs = {
            "s": self.advection_s(soln.w),
            "w": k ** 2 * self.invert_buoyancy(soln.s, soln.q),
            "q": self.advection_q(soln.w),
        }
        ins = soln._asdict()
        jac = dict_jacobian(outs, ins)
        jac = utils.torch2numpy(jac)
        jac = dict_to_matrix(jac, self.field_order)
        return jac

    @staticmethod
    def from_tom_base_state(tom_base_state):
        """Return WaveEq from Tom's base state"""
        # TODO refactor this to an abstract factor (A)
        tom_base_state = valmap(compose(np.copy, np.flip), tom_base_state)
        q, T, z, rho = get(["qv", "T", "z", "rho"], tom_base_state)
        base_state = {"QT": q * 1000.0, "SLI": T + grav / cp * z, "height_center": z, "density": rho}
        return WaveEq(base_state)


def tom_base_state_to_base_state(tom_base_state):
        tom_base_state = valmap(compose(np.copy, np.flip), tom_base_state)
        q, T, z, rho = get(["qv", "T", "z", "rho"], tom_base_state)
        return {"QT": q * 1000.0, "SLI": T + grav / cp * z, "height_center": z, "density": rho}


def tom_jacobian_to_jacobian(panes):
    jac = defaultdict(dict)

    def flip_axes_and_squeeze(x):
        return x[::-1, ::-1].squeeze()

    q_scale = Lc / 1000
    t_scale = cp

    jac["q"]["q"] = panes["q"]["q"]
    jac["q"]["s"] = panes["q"]["T"] / (q_scale / t_scale)
    jac["s"]["q"] = panes["T"]["q"] / (t_scale / q_scale)
    jac["s"]["s"] = panes["T"]["T"]

    jac = valmap(valmap(flip_axes_and_squeeze), jac)

    return jac


class NumpyEncoder(json.JSONEncoder):
    """From stackoverflow:
    https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


@dataclass
class LinearResponseFunction:
    """Linearized Response Function in (SLI, QT) space

    Attributes
    ----------
    panes
        dictionary of the linearized response function.
        panes[response][input][i,j] is the sensitivity of response[i] to
        input[j]. This data using the following physical units:

            panes['q']['q'] : g/kg / (g/kg) / s
            panes['q']['s'] : g/kg / K / s
            panes['s']['q'] : K / (g / kg) / s
            panes['s']['s'] : K / K / s

    """

    panes: Dict[str, Dict[str, np.array]]
    base_state: Dict[str, np.array]

    def iterpanes(self):
        for i in self.panes:
            for j in self.panes[i]:
                yield self.panes[i][j]

    def first_pane(self):
        first_row = first(self.panes.values())
        return first(first_row.values())

    def size(self):
        return self.first_pane().shape[0]

    def dump(self, file):
        return json.dump(asdict(self), file, cls=NumpyEncoder)

    @staticmethod
    def load(file):
        spec = json.load(file)
        panes = spec.pop("panes")

        new_panes = {}
        for i in panes:
            new_panes[i] = {}
            for j in panes[i]:
                new_panes[i][j] = np.asarray(panes[i][j])
        return LinearResponseFunction(new_panes, **spec)

    def to_array(self, field_order):
        jac = ensure_panes_present(self.panes, field_order)
        return dict_to_matrix(jac, field_order)


@dataclass
class FilteredLinearResponseFunction(LinearResponseFunction):

    threshold: float = 1.0
    replace: float = 0.0

    def to_array(self, field_order):
        A = super(FilteredLinearResponseFunction, self).to_array(field_order)
        return filter_small_eigenvalues(A, self.threshold, self.replace)



def ablate_upper_atmosphere(jacobian, boundaries):
    return _fill_zero_above_input_level(jacobian, boundaries)

@dataclass
class LowerAtmosLRF:
    """Linearized Response Function with no upper atmospheric sensitivity

    Attributes
    ----------
    boundaries
        dict of upper level boundaries
    lrf
        linear response function argument. Must have a `panes` attribute.
    """

    boundaries: Mapping
    pane: LinearResponseFunction

    def to_array(self, field_order):
        jacobian = ensure_panes_present(self.lrf.panes, field_order)
        jacobian = _fill_zero_above_input_level(jacobian, self.boundaries)
        return dict_to_matrix(jacobian, field_order)


@dataclass
class WaveCoupler:
    """Couple Linearized Response Function to Waves

    Attributes
    ----------
    wave : WaveEq
    lrf : LinearizedResponseFunction
    base_state : dict of numpy arrays
    """

    def __init__(self, wave: WaveEq, lrf: LinearResponseFunction):
        self.wave = wave
        self.lrf = lrf

    @property
    def base_state(self):
        return self.wave.base_state

    @property
    def field_order(self):
        return self.wave.field_order

    def system_matrix(self, k):
        return self.wave.system_matrix(k) + self.source_jacobian()

    def source_jacobian(self):
        return self.lrf.to_array(self.field_order)

    def plot_eigs_spectrum(self):
        k, As = self.matrices()
        eigs = np.linalg.eigvals(As) / k[:, np.newaxis]

        plt.plot(k, eigs.imag, ".")

    def get_eigen_pair(self, k):
        return np.linalg.eig(self.system_matrix(k))

    def matrices(self):
        k = np.r_[:64] / 1000e3
        As = [self.system_matrix(kk) for kk in k]
        return k, As

    def source_terms(self, vec):
        return self.source_jacobian() @ vec

    @staticmethod
    def from_tom_data(tom_data):
        # TODO refactor this to an abstract factor (A)
        base_state = tom_base_state_to_base_state(tom_data['base_state'])
        jacobian = tom_jacobian_to_jacobian(tom_data["jacobian"])
        wave_eq = WaveEq(base_state)
        lrf =  LinearResponseFunction(panes=jacobian, base_state=base_state)
        return WaveCoupler(wave_eq, lrf)


def base_from_xarray(mean):
    base_state = {}
    for key in ["SLI", "QT", "SOLIN", "SST"]:
        base_state[key] = utils.xarray2numpy(mean[key])

    base_state["density"] = utils.xarray2numpy(mean.rho)
    base_state["height_interface"] = interface_heights(mean.z)
    base_state["height_center"] = utils.xarray2numpy(mean.z)
    base_state["pressure"] = utils.xarray2numpy(mean.p)
    return base_state


def _expand_horiz_dims(d):
    out = {}
    for key in d:
        arr = d[key]
        n = arr.dim()
        if n == 1:
            out[key] = arr.unsqueeze(-1).unsqueeze(-1).unsqueeze(0)
        elif n == 0:
            out[key] = arr.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        else:
            raise NotImplementedError
    return out


@curry
def model_plus_damping(model, x, d0=1 / 86400.0):
    """Compute output of uwnet model with damping for vertical velocity

    Expands the appropriate dimensions of the inputs and outputs
    """

    d0 = 1 / 86400.0

    x = TensorDict(x)
    w = torch.zeros_like(x["SLI"])
    d = _expand_horiz_dims(x)
    inputs = TensorDict(d)
    outputs = model(inputs)
    outputs = outputs.apply(torch.squeeze)
    outputs["W"] = -w * d0

    return outputs


# TODO Delete this function if it isn't used
@curry
def marginize_model_over_solin(model, solin, inputs):
    """Marginalize out the diurnal cycle
    """
    outputs = []
    for sol in solin:
        inputs["SOLIN"] = torch.ones_like(inputs["SOLIN"]) * sol
        outputs.append(model(inputs))
    return sum(outputs) / len(outputs)



def plot_eigvals(A):
    vals = np.linalg.eigvals(A)
    plt.plot(vals.real, vals.imag, ".")
    plt.xlabel(r"$\Re$")
    plt.ylabel(r"$\Im$")


def plot_phase_speed(A, k):
    vals = np.linalg.eigvals(A)
    plt.plot(vals.real * 86400, vals.imag / k, ".")
    plt.xlabel("Growth rate (1/day)")
    plt.ylabel("Phase speed (m/s)")



"""Logging object used by training routine"""
import os

import attr

import torch
from datetime import datetime


def get_git_rev():
    import subprocess
    import uwnet
    root_dir = os.path.dirname(uwnet.__path__[0])
    out = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=root_dir)
    return out.decode().strip()


class DBLogger(object):
    def log_run(self, args, config):
        self.run_id = self.insert_run({
            'start_timestamp': datetime.utcnow(),
            'training_data': os.path.abspath(args.input),
            "config": config,
            "args": vars(args),
            'git': {
                'rev': get_git_rev()
            }
        })  # yapf: disable

    def log_batch(self, batch_info):
        batch_info['run_id'] = self.run_id
        self.insert_batch(batch_info)

    def log_epoch(self, i, lstm):
        path = f"{i}.pkl"
        torch.save({'epoch': i, 'dict': lstm.to_dict()}, path)
        self.insert_epoch({
            'epoch': i,
            'run_id': self.run_id,
            'model': os.path.abspath(path)
        })


@attr.s
class TinyDBLogger(DBLogger):
    """Class for logging data to a TinyDB"""
    path = attr.ib()

    @property
    def db(self):
        import tinydb
        return tinydb.TinyDB(self.path)

    # epoch_table = db.table('epochs')

    def insert_run(self, arg):
        run_table = self.db.table('runs')
        return run_table.insert(arg)

    def insert_batch(self, arg):
        table = self.db.table('batches')
        return table.insert(arg)

    def insert_epoch(self, arg):
        table = self.db.table('epochs')
        return table.insert(arg)


@attr.s
class MongoDBLogger(DBLogger):
    database = attr.ib(default='uwnet')

    @property
    def db(self):
        from pymongo import MongoClient
        return MongoClient()[self.database]

    def insert(self, collection, post):
        return self.db[collection].insert_one(post).inserted_id

    def insert_batch(self, post):
        """can use meta class for this, but not worth it"""
        return self.insert('batch', post)

    def insert_epoch(self, post):
        return self.insert('epoch', post)

    def insert_run(self, post):
        return self.insert('run', post)



import matplotlib.pyplot as plt
import numpy as np
import torch
import xarray as xr

from . import common
# from src.data import open_data
from .. import thermo
from wave import *


BOX_COLOR = "lightblue"

class paths:
    all = "../../nn/NNAll/20.pkl"
    lower = "../../nn/NNLowerDecayLR/20.pkl"
    nostab = "../../nn/NNLowerNoStabPenalty/20.pkl"


def sortbyvalue(eig):
    cp = eig.value.imag
    gr = eig.value.real
    permutation = cp * 100 + gr
    return eig.sortby(permutation)


def get_eigen_pair_xarray(wave, k):
    A = wave.system_matrix(k)
    lam, r = np.linalg.eig(A)
    return xr.Dataset(
        {"value": (["m"], lam), "vector": (["f", "m"], r)}, coords={"k": k}
    )


def compute_spectrum(wave, long_wave_km=40e6, short_wave_km=100e3) -> xr.Dataset:
    high_freq = 2 * np.pi / short_wave_km
    low_freq = 2 * np.pi / long_wave_km

    k = np.linspace(low_freq, high_freq, 100)
    eigs = [get_eigen_pair_xarray(wave, kk) for kk in k]
    return xr.concat(eigs, dim="k")


def plot_struct_x(eig):

    cp = eig.value.imag / eig.k
    targ = 20
    i = np.abs(cp - targ).argmin()
    eig = eig.isel(m=i)
    plot_struct_eig(eig)


def plot_struct_eig(eig):

    z = eig["z"]
    w, s, q = np.split(eig.vector, 3)
    fig, (a, b, c) = plt.subplots(1, 3, figsize=(10, 3), constrained_layout=True)

    a.set_title("W")
    im = plot_struct_2d(w.values, z, ax=a)
    plt.colorbar(im, ax=a, fraction=0.05)

    b.set_title("S")
    im = plot_struct_2d(s.values, z, ax=b)
    plt.colorbar(im, ax=b, fraction=0.05)

    c.set_title("Q")
    im = plot_struct_2d(q.values, z, ax=c)
    plt.colorbar(im, ax=c, fraction=0.05)

    cp = float(eig.value.imag / eig.k)
    gr = 86400 * float(eig.value.real)

    fig.suptitle(f"cp = {cp:.2f} m/s; gr = {gr:.2f} 1/d")


def plot_struct_eig_p(
    vec, sources, p, rho, w_range=(-1, 1), s_range=(-0.5, 0.5), q_range=(-0.5, 0.5)
):

    fig, axs = plt.subplots(
        1, 5, figsize=(8, 3.5), constrained_layout=True, sharey=True, sharex=True
    )

    axs[0].invert_yaxis()

    p = np.asarray(p)
    rho = np.asarray(rho)

    w, s, q = np.split(vec, 3)
    _, q1, q2 = np.split(sources * 86400, 3)

    # convert to mb/hr
    w = thermo.omega_from_w(w, rho) * 3600 / 100
    # normalize by maximum w
    scale = np.max(np.abs(w))
    w, s, q = w / scale, s / scale, q / scale

    def get_kwargs(range):
        return dict(
            cmap="RdBu_r",
            # levels=np.linspace(range[0], range[1], 11)
        )

    def plot_pane(ax, title, *args, **kwargs):
        im = plot_struct_2d(*args, ax=ax, **kwargs)
        cbar = plt.colorbar(im, ax=ax, location="bottom")
        cbar.ax.set_xticklabels(cbar.ax.get_xticklabels(), rotation=70)

        ax.set_title(title, loc='left')

    plot_pane(axs[0], r"a) $\omega$ (hPa/hr)", w, p, **get_kwargs(w_range))
    plot_pane(axs[1], "b) s (K)", s, p, **get_kwargs(s_range))
    plot_pane(axs[2], "c) q (g/kg)", q, p, **get_kwargs(q_range))
    plot_pane(axs[3], "d) Q1 (K/day)", q1, p, **get_kwargs(s_range))
    plot_pane(axs[4], "e) Q2 (g/kg/day)", q2, p, **get_kwargs(q_range))

    axs[0].set_ylabel("p (mb)")

    return fig


def most_unstable(eig, c=100):
    m = int(np.abs(eig.value - c).argmin())
    return eig.isel(m=m)


def add_boxes(ax):
    max_cp = 5
    max_gr = 0.0
    width = 1000
    height = 100

    kwargs = dict(color=BOX_COLOR, zorder=-1000.0)

    rect_right = plt.Rectangle((max_cp, max_gr), width, height, **kwargs)
    rect_left = plt.Rectangle((-max_cp-width, max_gr), width, height, **kwargs)

    ax.add_patch(rect_left)
    ax.add_patch(rect_right)

    ax.axhline(0.0, color='black', zorder=-500)


def scatter_spectra(eig, ax=None, symlogy=True, cbar=True, box=True, s=10):
    if ax is None:
        ax = plt.gca()

    cp = eig.value.imag / eig.k
    gr = eig.value.real * 86400

    cp, gr, k = [_.values.ravel() for _ in xr.broadcast(cp, gr, eig.k)]

    if box:
        add_boxes(ax)


    with plt.rc_context({'figure.dpi': 100}):
        im = ax.scatter(cp, gr, c=k, cmap="viridis", s=s, rasterized=True)
    ax.set_ylim(gr.min(), gr.max())
    if symlogy:
        ax.set_yscale("symlog", linthresh=0.1)
    # ax.set_xticks([-50,-25,0,25,50])
    ax.set_xlabel("Phase speed (m/s)")
    ax.set_ylabel("Growth rate (1/d)")

    return im


def get_spectra_for_path(path):
    wave, mean = get_coupler(path)
    eig = compute_spectrum(wave)
    eig["z"] = mean.z
    eig["p"] = mean.p
    return eig


def get_mean_data():
    return open_data("mean").isel(y=32)


def get_coupler(path, **kwargs):
    mean = get_mean_data()
    model = torch.load(path)
    src_original = model_plus_damping(model, d0=1 / 86400 / 5)
    coupler = WaveCoupler.from_xarray_and_source(mean, source=src_original, **kwargs)
    return coupler, mean


def get_data():
    #     eig_no_penalty = get_spectra_for_path("../models/280/5.pkl")
    eig_unsable = get_spectra_for_path(paths.all)
    eig = get_spectra_for_path(paths.lower)
    no_pen = get_spectra_for_path(paths.nostab)

    return (eig, eig_unsable, no_pen)


def plot_compare_stability_models(data=None):
    if data is None:
        data = get_data()
    stable, unstable, no_penalty = data
    fig, (a, b, c) = plt.subplots(
        1, 3, figsize=(common.width, 3), sharey=True, sharex=True
    )

    scatter_spectra(stable.isel(k=slice(0, 32)), ax=a)
    plt.ylim(-2, 1)
    a.set_title("a) NO upper atmospheric input;\n with penalty")

    scatter_spectra(unstable, ax=b)
    b.set_title("b) upper atmospheric input;\n with penalty")

    scatter_spectra(no_penalty, ax=c)
    c.set_title("c)No upper atmosphere\n no penalty")

    return data


def find_eig(values, k, phase_speed, growth_rate):
    query = -phase_speed * 1j * k + growth_rate / 86400
    closest_index = np.argmin(np.abs(values - query))
    return closest_index


def plot_structure(coupler, p, rho, k, phase_speed, growth_rate, **kwargs):
    # get closest eigenvector
    values, vectors = coupler.get_eigen_pair(k)

    # get the matching eigenvector and value
    closest_index = find_eig(values, k, phase_speed, growth_rate)
    vector = vectors[:, closest_index]
    value = values[closest_index]

    sources = coupler.source_terms(vector)

    # plot the eigenpair
    fig = plot_struct_eig_p(vector, sources, p, rho, **kwargs)
    cp = -value.imag / k
    gr = value.real * 86400
    wave_length = 2 * np.pi / k / 1000
    plt.suptitle(
        rf"$\sigma$ = {gr:.2f} 1/d; $c_p$ ={cp:.2f} m/s; $\lambda$={wave_length:.0f} km"
    )
    return fig


def get_linear_response_functions():
    wave_all = get_coupler(paths.all)[0].lrf
    wave_lower = get_coupler(paths.lower)[0].lrf
    wave_nostab = get_coupler(paths.nostab)[0].lrf

    return wave_all, wave_lower, wave_nostab


def plot_lrf_eigs(lrf, ax=None):
    variables = ["s", "q"]
    A = lrf.to_array(variables)
    values = np.linalg.eigvals(A) * 86400

    ax.plot(values.real, values.imag, ".")
    ax.grid()


def plot_compare_models_lrfs():
    lrfs = get_linear_response_functions()
    fig, axs = plt.subplots(3, 1, sharey=True, sharex=True, figsize=(4, 6))
    titles = ["All", "Lower", "LowerNoStab"]

    for k in range(len(axs)):
        plot_lrf_eigs(lrfs[k], axs[k])
        axs[k].set_title(titles[k])
        axs[k].set_ylabel(r"$\Im$ (1/day)")

    axs[-1].set_xlabel(r"$\Re $(1/day)")


def spectra_report(
    model_path, xlim_zoom=(-1, 1), structure_plots=((1.0, 0.2), (0, 0.9)), lrf_lid=None
):
    coupler, mean = get_coupler(model_path, lrf_lid=lrf_lid)
    eig = compute_spectrum(coupler)

    with plt.style.context("ggplot"):
        scatter_spectra(eig)
        plt.figure()
        scatter_spectra(eig, symlogy=False)
        plt.xlim(xlim_zoom)
        plt.ylim(bottom=0)

        for k, cp, gr in structure_plots:
            plt.figure()
            plot_structure(coupler, mean.p, mean.rho, k, phase_speed=cp, growth_rate=gr)


def plot_structure_from_path(model_path, structure=None, lrf_lid=None):
    coupler, mean = get_coupler(model_path, lrf_lid=lrf_lid)

    eig = compute_spectrum(coupler)
    k, cp, gr = structure
    return plot_structure(coupler, mean.p, mean.rho, k, phase_speed=cp, growth_rate=gr)


# TODO: refactor these plotting routines
def plot_struct_2d(w, z, n=256, ax=None, **kwargs):
    """Plot structure of eigenfunction over one phase of oscillation
    """
    phase = 2 * np.pi * np.r_[:n] / n
    phi = np.exp(1j * phase)[:, None]
    real_component = (w * phi).real
    im = ax.contourf(phase, z, real_component.T, **kwargs)
    ax.set_xlabel("phase (rad)")
    return im



#!/usr/bin/env python
"""
Package uwnet into a tarball for distribution

Usage:

python -m uwnet.archive 9_energy_conservation_change_loss/4.pkl uwnet.tar
"""
import sys
import sh
import os

model_file, output = sys.argv[1:]

git_files = sh.git('ls-files').split()
git_files.append(model_file)
git_files = [os.path.join(os.path.relpath(file, ".."))
             for file in git_files]

sh.tar('-C', '..', '-cvf', output, *git_files)



import json
from glob import glob

import dask.bag as db
import pandas as pd

import torch
import uwnet.loss
from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric


def r2_score(truth, pred, mean_dims, dims=None, w=1.0):
    """ R2 score for xarray objects
    """
    if dims is None:
        dims = mean_dims

    mu = truth.mean(mean_dims)
    sum_squares_error = ((truth - pred)**2 * w).sum(dims)
    sum_squares = ((truth - mu)**2 * w).sum(dims)

    return 1 - sum_squares_error / sum_squares


class WeightedMeanSquaredError(Metric):
    """
    Calculates the mean squared error.

    - `update` must receive output of the form `(y_pred, y)`.
    """

    def __init__(self, weights, *args, **kwargs):
        super(WeightedMeanSquaredError, self).__init__(*args, **kwargs)
        self.weights = weights.view(-1, 1, 1) / weights.sum()

    def reset(self):
        self._sum_of_squared_errors = 0.0
        self._num_examples = 0

    def update(self, output):
        y_pred, y = output
        squares = (y - y_pred)**2
        squared_errors = (squares * self.weights).sum()
        self._sum_of_squared_errors += squared_errors.item()
        self._num_examples += y.shape[0] * y.shape[1]

    def compute(self):
        if self._num_examples == 0:
            raise NotComputableError(
                'MeanSquaredError must have at least one example before it can be computed.'
            )
        return self._sum_of_squared_errors / self._num_examples


def read_metrics_files():
    """Read the metrics for all neural networks into a pandas dataframe"""
    jsons = glob('nn/**/*.json', recursive=True)
    json_seq = db.from_sequence(jsons)

    def read_json(path):
        with open(path) as f:
            d = json.load(f)
        d['path'] = path
        return d

    df = json_seq.map(read_json).to_dataframe().compute()

    model_info = df.path.str.extract("(?P<type>.*?)/(?P<model>.*?)/(?P<epoch>.*).json")
    df_with_info = pd.concat([df, model_info], axis=1)
    df_with_info['epoch'] = df_with_info.epoch.astype(int)

    return df_with_info.set_index(['type', 'model', 'epoch', 'path'])



import math

import numpy as np
import torch
from torch import nn

from .tensordict import TensorDict


class LinearDictIn(nn.Module):
    """A Dense Linear Layer Transform for dicts of tensors
    """

    def __init__(self, inputs, n_out, cls=nn.Linear):
        super(LinearDictIn, self).__init__()
        self.models = {}
        for field, num in inputs:
            self.models[field] = cls(num, n_out)
            # register submodule
            self.add_module(field, self.models[field])

        self.bias = nn.Parameter(torch.Tensor(n_out))
        self.reset_parameters()

    def forward(self, input):
        return sum(self.models[key](input[key])
                   for key in self.models) + self.bias

    def reset_parameters(self):
        n = sum(lin.weight.size(1) for lin in self.models.values())
        stdv = 1. / math.sqrt(n)
        for key in self.models:
            self.models[key].weight.data.uniform_(-stdv, stdv)

        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)


class LinearDictOut(nn.Module):
    """A Dense Linear Layer Transform for dicts of tensors
    """

    def __init__(self, n_in, outputs, cls=nn.Linear):
        super(LinearDictOut, self).__init__()
        self.models = {}
        for field, num in outputs:
            self.models[field] = cls(n_in, num)
            # register submodule
            self.add_module(field, self.models[field])

    def forward(self, input):
        return TensorDict(
            {key: self.models[key](input)
             for key in self.models})


def get_affine_transforms(func, n):
    """Get weights matrix and bias of an affine function

    func(I) = A + b
    func(0) = b

    """
    identity = np.eye(n)
    z = np.zeros((1, n))

    b = func(z)
    A = func(identity) - b
    return A, b


class LinearFixed(nn.Module):
    """A linear tranform with fixed weights and bias

    Useful for representing scikit-learn affine transform functions
    """

    def __init__(self, *args):
        super(LinearFixed, self).__init__()

        for arg in args:
            arg.requires_grad = False

        weight, bias = args
        self.register_buffer('weight', weight)
        self.register_buffer('bias', bias)

    def forward(self, x: torch.tensor):
        return torch.matmul(x, self.weight) + self.bias

    @property
    def in_features(self):
        return self.weight.size(0)

    @property
    def out_features(self):
        return self.weight.size(1)

    @classmethod
    def from_affine(cls, func, n):
        """Initialize from an arbitrary python function

        Parameters
        ----------
        func
            function has input dimension n and output dimension m
        n : int
            input dimensions
        """
        args = get_affine_transforms(func, n)
        args = [torch.tensor(arg, requires_grad=True) for arg in args]
        return cls(*args)

    def __repr__(self):
        return f"LinearFixed({self.in_features}, {self.out_features})"


class MOE(nn.Module):
    def __init__(self, m, n, n_experts):
        "docstring"
        super(MOE, self).__init__()

        self.experts = nn.ModuleList(
            [nn.Sequential(nn.Linear(m, n), ) for _ in range(n_experts)])

        self.decider = nn.Sequential(
            nn.Linear(m, 256),
            nn.ReLU(),
            nn.Linear(256, n_experts),
            nn.Softmax(dim=1))

    def poll_experts(self, x):
        return [expert(x) for expert in self.experts]

    def forward(self, x):
        weights = self.decider(x)
        ans = 0
        for val, w in zip(self.poll_experts(x), weights.split(1, dim=-1)):
            ans = ans + w * val
        return ans


class _Partial(nn.Module):
    after = True

    def __init__(self, fun, *args, **kwargs):
        "docstring"
        super(_Partial, self).__init__()
        self.fun = fun
        self.args = args
        self.kwargs = kwargs

    def forward(self, *args, **kwargs):
        if self.after:
            fargs = args + self.args
        else:
            fargs = self.args + args

        fkwargs = self.kwargs.copy()
        fkwargs.update(kwargs)
        return self.fun(*fargs, **fkwargs)


class Partial(_Partial):
    """Module similar to functools.partial

    This allows a function to be easily wrapped as a torch module. Which makes
    it compatible with nn.Sequential and easy to save/load.
    """
    after = False


class RPartial(_Partial):
    """Same as Partial but stored args are passed after the first argument"""
    after = True


class ValMap(nn.Module):
    def __init__(self, fun):
        "docstring"
        super(ValMap, self).__init__()
        self.fun = fun

    def forward(self, x):
        from .tensordict import TensorDict
        out = {}
        for key, val in x.items():
            out[key] = self.fun(val)
        return TensorDict(out)


def mapbykey(funcs, d):
    return {key: funcs[key](d[key]) for key in funcs}


class MapByKey(nn.Module):
    """Apply modules to inputs by key

    This code explains the how this module works::

        {key: funcs[key](d[key]) for key in funcs}

    Examples
    --------
    >>> a = torch.ones(1)
    >>> mapbykey({'a': lambda x: x}, {'a': a})
    """

    def __init__(self, funcs):
        super(MapByKey, self).__init__()
        self.funcs = nn.ModuleDict(funcs)

    def __getitem__(self, key):
        return self.funcs[key]

    @property
    def outputs(self):
        return [(key, self.funcs[key].out_features) for key in self.funcs]

    @property
    def inputs(self):
        return [(key, self.funcs[key].in_features) for key in self.funcs]

    def forward(self, d):
        return TensorDict(mapbykey(self.funcs, d))


class ConcatenatedWithIndex(nn.Module):
    """Module for concatenating the output of another module in a reproducable
    order"""
    def __init__(self, model, dim=-1):
        "docstring"
        super(ConcatenatedWithIndex, self).__init__()
        self.model = model
        self.dim = dim

        keys = []
        index = []
        for key, n in self.model.outputs:
            keys.append(key)
            for k in range(n):
                index.append(k)

        self.keys = keys
        self.index = torch.tensor(index)

    def forward(self, x):
        output = []
        y = self.model(x)
        for key in self.keys:
            output.append(y[key])
        return torch.concatenate(output, self.dim)



"""Thermodynamic and other math calculations
"""
from functools import partial
import numpy as np
import xarray as xr
from .xcalc import centderiv

grav = 9.81
R = 287.058
cp = 1004
kappa = R / cp
Lc = 2.5104e6
rho0 = 1.19
sec_in_day = 86400
liquid_water_density = 1000.0

rad_earth = 6371e3  # m
circumference_earth = rad_earth * 2 * np.pi


def compute_insolation(lat, day, scon=1367, eccf=1.0):
    """Compute the solar insolation in W/m2 assuming perpetual equinox

    Parameters
    ----------
    lat : (ny, nx)
        latitude in degrees
    day : float
        day of year. Only uses time of day (the fraction).
    scon : float
        solar constant. Default 1367 W/m2
        eccentricity factor. Ratio of orbital radius at perihelion and
        aphelion. Default 1.0.

    """
    time_of_day = day % 1.0

    # cos zenith angle
    mu = -np.cos(2 * np.pi * time_of_day) * np.cos(np.pi * lat / 180)
    mu[mu < 0] = 0.0
    return scon * eccf * mu

def metpy_wrapper(fun):
    """Given a metpy function return an xarray compatible version
    """
    from metpy.units import units as u

    def func(*args):
        def f(*largs):
            new_args = [
                u.Quantity(larg, arg.units) for larg, arg in zip(largs, args)
            ]
            return fun(*new_args)

        output_units = f(* [1 for arg in args]).units
        ds = xr.apply_ufunc(f, *args)
        ds.attrs['units'] = str(output_units)
        return ds

    return func


def omega_from_w(w, rho):
    """Presure velocity in anelastic framework

    omega = dp_0/dt = dp_0/dz dz/dt = - rho_0 g w
    """
    return -w * rho * grav


def liquid_water_temperature(t, qn, qp):
    """This is an approximate calculation neglecting ice and snow
    """

    sl = t + grav / cp * t.z - Lc / cp * (qp + qn) / 1000.0
    sl.attrs['units'] = 'K'
    return sl


def total_water(qv, qn):
    qt = qv + qn
    qt.attrs['units'] = 'g/kg'
    return qt


def get_dz(z):
    zext = np.hstack((-z[0], z, 2.0 * z[-1] - 1.0 * z[-2]))
    zw = .5 * (zext[1:] + zext[:-1])
    dz = zw[1:] - zw[:-1]

    return xr.DataArray(dz, z.coords)


def interface_heights(z):
    zext = np.hstack((-z[0], z, 2.0 * z[-1] - 1.0 * z[-2]))
    return .5 * (zext[1:] + zext[:-1])


def layer_mass(rho):
    dz = get_dz(rho.z)
    return (rho * dz).assign_attrs(units='kg/m2')


def layer_mass_from_p(p, ps=None):
    if ps is None:
        ps = 2 * p[0] - p[1]

    ptop = p[-1] * 2 - p[-2]

    pext = np.hstack((ps, p, ptop))
    pint = (pext[1:] + pext[:-1]) / 2
    dp = -np.diff(pint * 100) / grav

    return xr.DataArray(dp, p.coords)


def mass_integrate(p, x, average=False):
    dp = layer_mass_from_p(p)
    ans = (x * dp).sum(p.dims)

    if average:
        ans = ans / dp.sum()

    return ans


def column_rh(QV, TABS, p):
    from metpy.calc import relative_humidity_from_mixing_ratio
    rh = metpy_wrapper(relative_humidity_from_mixing_ratio)(QV, TABS, p)

    return mass_integrate(p, rh / 1000, average=True)


def ngaqua_y_to_lat(y, y0=5120000):
    rad_earth = 6371e3  # m
    circumference_earth = rad_earth * 2 * np.pi
    return (y - y0) / circumference_earth * 360


def coriolis_ngaqua(y):
    lat = ngaqua_y_to_lat(y)
    omega = 2 * np.pi / 86400
    return 2 * omega * np.sin(np.deg2rad(lat))


def get_geostrophic_winds(p, rho, min_cor=1e-5):
    """Compute geostrophic winds

    Parameters
    ----------
    p : xr.DataArray
        partial pressure in (Pa)
    rho : xr.DataArray
        density in (kg/m3)
    min_cor : float
        minimum coriolis paramter

    Returns
    -------
    ug, vg : xr.DataArray
        the geostropohc wind fields, with values masked for locations where the
        absolute coriolis parameter is smaller than min_cor.

    """
    # get coriolis force
    fcor = coriolis_ngaqua(p.y)
    px = centderiv(p, dim='x') / rho
    py = centderiv(p, dim='y') / rho

    vg = px / fcor
    vg = vg.where(np.abs(fcor) > min_cor)
    vg.name = "VG"

    ug = -py / fcor
    ug = ug.where(np.abs(fcor) > min_cor)
    ug.name = "UG"
    return ug, vg


def compute_apparent_source(prog, forcing):
    dt = prog.time[1] - prog.time[0]
    avg_forcing = (forcing + forcing.shift(time=-1)) / 2
    return (prog.shift(time=-1) - prog) / dt - avg_forcing


def compute_q2(ngaqua):
    return compute_apparent_source(ngaqua.QT, ngaqua.FQT * 86400)


def vorticity(u, v):
    psi = u.differentiate('y') - v.differentiate('x')
    psi.name = 'Vorticity'
    return psi


def lhf_to_evap(lhf):
    rhow = 1000
    evap = lhf / 2.51e6 / rhow * 86400 * 1000
    evap.name = 'Evaporation'
    evap.attrs['units'] = 'mm/day'
    return evap


def integrate_q1(q1, layer_mass, dim='z'):
    """Vertically integrate Q1 (K/day) to give W/m2
    """
    return (q1 * layer_mass).sum(dim) * (cp / sec_in_day)


def integrate_q2(q2, layer_mass, dim='z'):
    """Vertically integrate Q2 (g/kg/day) to give mm/day
    """
    return (q2 * layer_mass).sum(dim) / liquid_water_density


def net_precipitation_from_training(data):
    """Compute Net Precipitation from Q2

    This is not exactly equivalent to precipitation minus evaporation due to
    sampling issue.
    """
    return -integrate_q2(
        compute_apparent_source(data.QT, data.FQT * 86400), data.layer_mass)


def net_precipitation_from_prec_evap(data):
    return data.Prec - lhf_to_evap(data.LHF)


def net_heating(prec, shf, swns, swnt, lwns, lwnt):
    surface_radiation_net_upward = (lwns - swns)
    toa_radiation_net_upward = (lwnt - swnt)
    net_radiation = surface_radiation_net_upward - toa_radiation_net_upward

    return prec * (Lc / sec_in_day) + net_radiation


def net_heating_from_data_2d(data_2d):
    prec = data_2d.Prec
    shf = data_2d.SHF
    swns = data_2d.SWNS
    swnt = data_2d.SWNT
    lwns = data_2d.LWNS
    lwnt = data_2d.LWNT
    return net_heating(prec, shf, swns, swnt, lwns, lwnt)


def periodogram(pw: xr.DataArray, dim='x', freq_name='f'):
    from scipy import signal
    axis = pw.get_axis_num(dim)

    x = pw.values

    coord = pw[dim]
    d = float(coord[1]-coord[0])
    f, x = signal.periodogram(x, axis=axis, fs=1/d)

    dims = list(pw.dims)
    dims[pw.get_axis_num(dim)] = freq_name
    coords = {key: pw[key] for key in pw.dims if key != dim}
    coords[freq_name] = f

    return xr.DataArray(x, dims=dims, coords=coords)


def water_budget(data_2d):
    """Compute precipitable water budget from 2D data"""
    storage = data_2d.PW.differentiate('time')
    advection = storage + data_2d.NPNN
    return xr.Dataset({'storage': storage, 'advection': advection, 'net_precip':  data_2d.NPNN})


def potential_temperature(temperature_kelvin, pressure_mb, p0=1015.0):
    return temperature_kelvin * (p0 / pressure_mb)**kappa


def lower_tropospheric_stability(
        temperature_kelvin, pressure_mb, sst, p0=1015.0):
    theta = potential_temperature(temperature_kelvin, pressure_mb, p0)
    i = int(np.argmin(np.abs(np.asarray(pressure_mb) - 700)))
    return theta.isel(z=i) - sst


def water_vapor_path(qv, p, bottom=850, top=550, dim='z'):
    """Water vapor path between specified pressure levels

    Parameters
    ----------
    qv
        water vapor in g/kg
    p
        pressure in mb
    bottom, top : float
        pressure at bottom and top
    dim : default 'z'
        vertical dimension of the data

    Returns
    -------
    path
        water vapor path in mm (liquid water equivalent)

    """

    dp = layer_mass_from_p(p)
    masked = qv.where((p < bottom) & (p > top), 0)
    mass = (masked * dp).sum(dim=dim)
    return mass/liquid_water_density


def layer_mass_from_p(p, ps=None):
    if ps is None:
        ps = 2 * p[0] - p[1]

    ptop = p[-1] * 2 - p[-2]

    pext = np.hstack((ps, p, ptop))
    pint = (pext[1:] + pext[:-1]) / 2
    dp = -np.diff(pint * 100) / grav

    return xr.DataArray(dp, p.coords)


midtropospheric_moisture = partial(water_vapor_path, bottom=850, top=600)


def omega_from_w(w, rho):
    """Presure velocity in anelastic framework

    omega = dp_0/dt = dp_0/dz dz/dt = - rho_0 g w
    """
    return -w * rho * grav


import attr
import click
import torch

import xarray as xr
from uwnet.timestepper import Batch, predict_multiple_steps


def _convert_dataset_to_dict(dataset):
    return {
        key: dataset[key]
        for key in dataset.data_vars if 'time' in dataset[key].dims
    }


class XarrayBatch(Batch):
    """An Xarray-aware version of batch"""

    def __init__(self, dataset, **kwargs):
        self._dataset = dataset
        data = _convert_dataset_to_dict(dataset)
        super(XarrayBatch, self).__init__(data, **kwargs)

    @staticmethod
    def select_time(data, t):
        return data.apply(lambda x: x.isel(time=t))

    def get_model_inputs(self, t, state):
        inputs = super(XarrayBatch, self).get_model_inputs(t, state)
        for key in inputs:
            try:
                inputs[key] = inputs[key].drop('time')
            except ValueError:
                pass
        return xr.Dataset(inputs)

    def get_prognostics_at_time(self, t):
        return self._dataset[self.prognostics].isel(time=t)


def _get_time_step(ds):
    return float(ds.time.diff('time')[0] * 86400)


def single_column_simulation(model,
                             dataset,
                             interval=None,
                             prognostics=(),
                             time_step=None):
    """Run a single column model simulation with a model for the source terms

    Parameters
    ----------
    model
        pytorch model for producing the apparent sources
    dataset : xr.Dataset
        input dataset in the same format as the training data
    interval : tuple
        (start_time, end_time) interval
    """
    if not time_step:
        time_step = _get_time_step(dataset)

    if not interval:
        start, end = 0, len(dataset.time) - 1
    else:
        start, end = interval

    batch = XarrayBatch(dataset, prognostics=prognostics)
    pred_generator = predict_multiple_steps(
        model.call_with_xr,
        batch,
        initial_time=start,
        prediction_length=end - start,
        time_step=time_step)
    datasets = []
    for k, state, diag in pred_generator:

        for key in diag:
            if key in state:
                diag = diag.rename({key: 'F' + key + 'NN'})

        datasets.append(xr.Dataset(state).assign_coords(time=dataset.time[k]).merge(diag))
    output_time_series = xr.concat(datasets, dim='time')
    return output_time_series


def compute_apparent_sources(model, ds):
    sources = model.call_with_xr(ds)
    rename_dict = {}
    for key in sources.data_vars:
        rename_dict[key] = 'F' + key + 'NN'
    return sources.rename(rename_dict)


def remove_nonphysical_dims(data):
    """Remove all dimensions other than x y z or time"""
    true_dims = ['x', 'y', 'z', 'time']
    for dim in data.dims:
        if dim not in true_dims:
            data = data.isel(**{dim: 0})
    return data


@click.command()
@click.argument('model')
@click.argument('data')
@click.argument('output_path')
@click.option('-b', '--begin', type=int)
@click.option('-e', '--end', type=int)
def main(model, data, output_path, begin, end):
    model = torch.load(model)
    data = xr.open_dataset(data)

    if begin is None:
        begin = 0
    if end is None:
        end = len(data.time)

    data = data.isel(time=slice(begin, end))
    data = remove_nonphysical_dims(data)
    output = single_column_simulation(model, data)
    sources = compute_apparent_sources(model, data)
    output.merge(sources).to_netcdf(output_path)


if __name__ == '__main__':
    main()



"""Calculus for xarray data
"""
import numpy as np
import xarray as xr

import dask.array as da


def dask_centdiff(x, axis=-1, boundary='reflect'):
    """Compute centered difference using dask
    """

    def fun(x):
        return np.roll(x, -1, axis) - np.roll(x, 1, axis)

    return da.overlap.map_overlap(x, fun, {axis: 1},
                                boundary={axis: boundary},
                                dtype=x.dtype)


def centdiff(A, dim='x', boundary='periodic'):
    if A.chunks is None:
        A = A.chunk()

    return xr.apply_ufunc(
        dask_centdiff,
        A,
        input_core_dims=[['x']],
        output_core_dims=[['x']],
        dask='allowed',
        kwargs={"boundary": boundary}
    )


def centspacing(x):
    """Return spacing of a given xarray coord object
    """

    dx = x.copy()

    x = x.values

    xpad = np.r_[2*x[0]-x[1], x, 2* x[-1] - x[-2]]

    dx = xr.DataArray(xpad[2:] - xpad[:-2], dims=dx.dims)

    return dx


def centderiv(A, dim='x', boundary='periodic'):
    """Compute centered derivative of a data array

    Parameters
    ----------
    A: DataArray
        input dataset
    dim: str
        dimension to take derivative along
    boundary: str
        boundary conditions along dimension. One of 'periodic', 'reflect',
        'nearest', 'none', or an value.

    See Also
    --------
    dask.array.ghost.ghost

    """
    return centdiff(A, dim=dim, boundary=boundary)/centspacing(A[dim])


# monkey patch DataArray class
xr.DataArray.centderiv = centderiv



import xarray as xr

# needed for arr.ndimage xarray acessor
from .. import ndimage_xarray


def blur(arr, sigma):
    if not {'x', 'y'} <= set(arr.dims):
        raise ValueError
    if sigma <= 1e-8:
        raise ValueError

    return (arr.ndimage.gaussian_filter1d('x', sigma, mode='wrap')
            .ndimage.gaussian_filter1d('y', sigma, mode='nearest'))


def blur_dataset(data, sigma):
    data_vars = {}
    for x in data:
        try:
            data_vars[x] = blur(data[x], sigma)
        except ValueError:
            data_vars[x] = data[x]

    return xr.Dataset(data_vars)



from functools import partial

import torch
import xarray as xr
from toolz import pipe
from .tensordict import TensorDict


def _is_at_least_2d(da):
    return set(da.dims) >= {'x', 'y'}


def _array_to_tensor(da):
    if _is_at_least_2d(da):
        if 'z' not in da.dims:
            da = da.expand_dims('z', -3)
    return torch.from_numpy(da.values)\
                .detach()\
                .float()


def dataset_to_torch_dict(ds):
    return TensorDict({
        key: _array_to_tensor(ds[key])
        for key in ds.data_vars
    })


def _torch_dict_to_dataset(output, coords):
    # parse output
    data_vars = {}

    dims_2d = [dim for dim in ['time', 'y', 'x'] if dim in coords.dims]
    dims_3d = [dim for dim in ['time', 'z', 'y', 'x'] if dim in coords.dims]

    time_dim_present = 'time' in dims_2d

    for key, val in output.items():
        z_dim = 1 if time_dim_present else 0
        nz = val.size(z_dim)

        if nz == 1:
            data_vars[key] = (dims_2d, output[key].numpy().squeeze(z_dim))
        else:
            data_vars[key] = (dims_3d, output[key].numpy())

    # prepare coordinates
    coords = dict(coords.items())
    return xr.Dataset(data_vars, coords=coords)


def _assert_no_null_dimensions(ds):
    for key in ds.data_vars:
        for k, n in enumerate(ds[key].shape):
            if ds[key].shape[k] == 0:
                dim = ds[key].dims[k]
                raise ValueError(
                    f"'{dim}' dimension of '{key}' has length 0"
                )


def call_with_xr(self, ds, **kwargs):
    """Call the neural network with xarray inputs"""
    _assert_no_null_dimensions(ds)
    tensordict = dataset_to_torch_dict(ds)
    with torch.no_grad():
        output = self(tensordict, **kwargs)
    return _torch_dict_to_dataset(output, ds.coords)


class XRCallMixin(object):
    """PyTorch module for predicting Q1, Q2 and maybe Q3"""
    call_with_xr = call_with_xr
    predict = call_with_xr


class XarrayWrapper(object):
    def __init__(self, model):
        self.model = model

    def __call__(self, *args, **kwargs):
        return call_with_xr(self.model, *args, **kwargs)



import numpy as np
import xarray as xr


def assert_tensors_allclose(*args, atol=1e-7, **kwargs):
    args = [arg.detach().numpy() for arg in args]
    return np.testing.assert_allclose(*args, atol=atol, **kwargs)


def mock_data(shape=(4, 5, 6, 7), init=np.zeros):
    dims = ['time', 'y', 'x', 'z']
    return xr.DataArray(init(shape), dims=dims)



import matplotlib.pyplot as plt
import xarray as xr
import click
import torch
from uwnet.model import call_with_xr, model_factory


def column_integrate(data_array, mass):
    return (data_array * mass).sum('z')

def plot_water_imbalance(ds, mod, max_height, ax=None):

    if max_height:
        ds = ds.isel(z=slice(max_height))

    output = call_with_xr(mod, ds, drop_times=0)


    net_precip = -column_integrate(output.QT, ds.layer_mass).mean(['x', 'time']) / 1000
    cfqt = column_integrate(ds.FQT, ds.layer_mass).mean(['x', 'time']) * 86400 / 1000

    # plot data
    cfqt.plot(label='cFQT', ax=ax)
    net_precip.plot(label='Net Precipitation', ax=ax)
    ax.autoscale('y', tight=True)
    ax.legend()
    ax.set_ylabel('mm/day')


@click.command()
@click.argument('dataset')
@click.argument('model')
@click.option('-z', type=int, default=None, help='maximum height used in the network')
@click.option('-o', '--output-path', type=click.Path(), default=None, help='where to save figure')
def main(dataset, model, z, output_path=None):
    ds = xr.open_dataset(dataset).isel(time=slice(0, 120))
    mod = torch.load(model)
    fig, ax = plt.subplots()
    plot_water_imbalance(ds, mod, z, ax=ax)
    if output_path:
        plt.savefig(output_path)
    else:
        plt.show()


if __name__ == '__main__':
    main()



import torch
from uwnet.loss import get_input_output
from uwnet.utils import mean_other_dims
import json
from tqdm import tqdm
import argparse

import uwnet.ml_models.nn.datasets_handler as d


def batch_to_residual(model, batch):
    from uwnet.timestepper import Batch

    batch = Batch(batch.float(), prognostics=["QT", "SLI"])
    with torch.no_grad():
        prediction, truth = get_input_output(model, 0.125, batch)
    return prediction - truth


def vertically_resolved_mse_from_residual(residual):
    return {k: mean_other_dims(residual[k] ** 2, 2).squeeze() for k in residual}


def batch_to_mse(model, batch):
    residual = batch_to_residual(model, batch)
    return vertically_resolved_mse_from_residual(residual)


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("data", help="path to zarr reshaped training or test data")
    parser.add_argument("model", help="path to model")

    return parser.parse_args()


args = _parse_args()

model_path = args.model
path = args.data
prognostics = ["QT", "SLI"]

model = torch.load(model_path)
train_dataset = d.get_dataset(path, predict_radiation=False)
dl = d.get_data_loader(train_dataset, prognostics, batch_size=64)

total = {}
count = 0
for batch in tqdm(dl):
    mse = batch_to_mse(model, batch)
    for key in mse:
        count += 1
        alpha = 1 / count
        zeros = torch.zeros_like(mse[key])
        total[key] = total.get(key, zeros) * (1 - alpha) + mse[key] * alpha

total = {"mse_apparent_source_" + k.lower(): total[k].numpy().tolist() for k in total}
print(json.dumps(total))



"""Interface for calling neural network models from SAM

Configuration
-------------

This interface routine is configured using environmental variables. Currently
these are

- UWNET_MODEL : The path to the model in a pickle file


Nudging
~~~~~~~

UWNET_NUDGE_TIME_SCALE: nudging time-scale
NGAQUA_PATH : path to target ngaqua data

"""
# import debug
import logging
import os

import numpy as np
from toolz import curry, valmap

import torch
from torch import nn
from uwnet.numpy_interface import NumpyWrapper
from uwnet.sam_ngaqua import get_ngaqua_nudger
import json
from cProfile import Profile


profiler = Profile()


def initialize(state):
    profiler.enable()


def finalize(state):
    profiler.disable()
    profiler.dump_stats("python.profile")

from uwnet.thermo import compute_insolation


def get_configuration():
    try:
        with open("python_config.json") as f:
            return json.load(f)
    except FileNotFoundError:
        return {'models': []}


def rename_keys(rename_table, d):
    return {rename_table.get(key, key): d[key] for key in d}


@curry
def CFVariableNameAdapter(model, d, label=''):
    """Wrapper for translating input/output variable names in the neural network
    model to CF-compliant ones"""

    table = [
        ("liquid_ice_static_energy", "SLI"),
        ("x_wind", "U"),
        ("y_wind", "V"),
        ("upward_air_velocity", "W"),
        ("total_water_mixing_ratio", "QT"),
        ("air_temperature", "TABS"),
        ("latitude", "lat"),
        ("longitude", "lon"),
        ("sea_surface_temperature", "SST"),
        ("surface_air_pressure", "p0"),
        ("toa_incoming_shortwave_flux", "SOLIN"),
        ("surface_upward_sensible_heat_flux", "SHF"),
        ("surface_upward_latent_heat_flux", "LHF"),
    ]

    input_keys = dict(table)
    output_keys = dict((y, x) for x, y in table)

    # Translate from CF input names to the QT, SLI, names
    d_with_old_names = rename_keys(input_keys, d)
    out = model(d_with_old_names)

    # translate output keys to CF
    out = rename_keys(output_keys, out)
    # add tendency_of to these keys
    tendency_names = {
        key: 'tendency_of_' + key + '_due_to_' + label
        for key in out
    }

    return rename_keys(tendency_names, out)


def get_model(config):
    type = config['type']
    if type == 'neural_network':
        model = torch.load(config['path'])
        if isinstance(model, nn.Module):
            model.eval()
            model = NumpyWrapper(model)
        return CFVariableNameAdapter(model, label='neural_network')
    elif type == "cf":
        return torch.load(config['path'])
    elif type == 'nudging':
        return CFVariableNameAdapter(
            get_ngaqua_nudger(config), label='nudging')
    else:
        raise NotImplementedError(f"Model type {type} not implemented")


CONFIG = get_configuration()
MODELS = [get_model(model) for model in CONFIG['models']]


def sum_up_tendencies(d):
    import re
    output = {}

    pattern = re.compile('tendency_of_(.*)_due_to')

    for key in d:
        m = pattern.search(key)
        if m:
            variable_name = 'tendency_of_' + m.group(1)
        else:
            variable_name = key

        seq = output.setdefault(variable_name, [])
        seq.append(d[key])

    return valmap(sum,  output)


def call_neural_network(state):

    logger = logging.getLogger(__name__)

    # compute insolation. it is difficult to configure SAM to compute this
    # without including radiation
    state['SOLIN'] = compute_insolation(state['latitude'], state['day'])

    # Compute the output of all the models
    all_outputs = {}
    for model in MODELS:
        out = model(state)
        all_outputs.update(out)

    totaled_output = sum_up_tendencies(all_outputs)

    # update the state
    state.update(totaled_output)
    state.update(all_outputs)


def call_save_state(state):
    """This simple function can be called in SAM using the following namelist
    entries:

    module_name = <this modules name>,
    function_name = 'call_save_state'

    """
    step = int(state.get('nstep', 0))
    caseid = state['caseid']
    case = state['case']
    name = f"OUT_3D/{case}_{caseid}_{step:010d}.pt"
    torch.save(state, name)



"""SAM Python plugins for forcing and nudging coarse resolution SAM towards
NG-Aqua"""
import os
import torch
import logging
import xarray as xr
import numpy as np
from toolz import valmap, curry
from uwnet.thermo import compute_apparent_source
from functools import lru_cache


def _dataarray_to_numpy(x):
    dim_order = ['z', 'y', 'x']
    return x.transpose(*dim_order).values


@curry
def _get_numpy_from_time(x, time):
    return x.sel(time=time, method='nearest').pipe(_dataarray_to_numpy)


class NGAquaForcing(object):
    def __init__(self, ngaqua):
        self.logger = logging.getLogger('NGAquaForcing')
        self.logger.info("Compute NGAqua Forcing from time series")

        self.sources = dict(
            FQTNN=compute_apparent_source(ngaqua.QT, ngaqua.FQT * 86400),
            FSLINN=compute_apparent_source(ngaqua.SLI, ngaqua.FSLI * 86400))

    def get_forcings(self, time):
        self.logger.info(f"Getting forcing from day {time}")
        return valmap(_get_numpy_from_time(time=time), self.sources)


class NGAquaNudger(object):
    """Class for nudging SAM towards NG-Aqua"""

    def __init__(self,
                 ngaqua,
                 time_scale=.125,
                 nudging_variables=('QT', 'SLI', 'U', 'V')):
        self.logger = logging.getLogger('NGAquaNudger')
        self.ngaqua = ngaqua
        self.time_scale = time_scale
        self.nudging_variables = nudging_variables

    def get_previous_time_index(self, day):
        return np.searchsorted(self.ngaqua.time.values, day, side='right') - 1

    @lru_cache(maxsize=20)
    def cached_get_numpy_at_time(self, field, time_index):
        return _dataarray_to_numpy(self.ngaqua[field].isel(time=time_index))

    def get_ngaqua_at_day(self, day):
        before = self.get_previous_time_index(day)
        after = before + 1

        t0 = float(self.ngaqua.time[before])
        t1 = float(self.ngaqua.time[after])

        dt = t1 - t0
        a = (day - t0) / dt
        assert 0 <= a < 1

        target = {}
        for key in self.nudging_variables:
            n0 = self.cached_get_numpy_at_time(key, before)
            n1 = self.cached_get_numpy_at_time(key, after)

            target[key] = (1 - a) * n0 + a * n1
        return target

    def get_nudging(self, state):
        day = state['day']
        self.logger.info(f"Nudging at day {day}")
        target = self.get_ngaqua_at_day(day)
        return self.nudge_variables(state, target)

    def nudge_variable(self, x, target):
        return (target - x) / self.time_scale

    def _get_out_key(self, key):
        return key

    def nudge_variables(self, state, target):
        output = {}
        for key in self.nudging_variables:
            out_key = self._get_out_key(key)
            src = self.nudge_variable(state[key], target[key])
            output[out_key] = src
        return output

    def __call__(self, *args):
        return self.get_nudging(*args)


def get_ngaqua_forcing():
    path = os.environ['NGAQUA_PATH']
    ds = xr.open_dataset(path).isel(step=0).chunk({'time': 1})
    return NGAquaForcing(ds)


def get_ngaqua_nudger(config):
    path = config['ngaqua']
    time_scale = config['time_scale']
    ds = xr.open_dataset(path).chunk({'time': 1})
    return NGAquaNudger(ds, time_scale=time_scale,
                        nudging_variables=['U', 'V', 'W', 'SLI', 'QT'])



import logging

import attr
import matplotlib.pyplot as plt

import xarray as xr

from .columns import single_column_simulation


@attr.s
class TrainingPlotManager(object):
    """Manages the creation of plots during the training process"""

    experiment = attr.ib()
    model = attr.ib()
    dataset = attr.ib()
    single_column_locations = attr.ib()
    interval = attr.ib(default=1)

    def __call__(self, engine):
        return self.plot(engine)

    def plot(self, engine):
        n = engine.state.epoch

        if n % self.interval != 0:
            return

        ex = self.experiment
        imbalance_plot(self.model, self.dataset, engine)
        single_column_plots = [plot_q2(ex), plot_scatter_q2_fqt(ex)]
        for y, x in self.single_column_locations:
            location = self.dataset.isel(
                y=slice(y, y + 1), x=slice(x, x + 1), time=slice(0, 200))

            try:
                output = self.model.call_with_xr(location)
            except ValueError:
                continue

            for plot in single_column_plots:
                plot.save_figure(f'{engine.state.epoch}-{y}', location, output)

            i = engine.state.epoch
            filenames = [
                name + f'{i}-{y}'
                for name in ['qt', 'fqtnn', 'fqtnn-obs', 'pw']
            ]
            # water_budget_plots(self.model, self.dataset, location, filenames)


def water_budget_plots(model, ds, location, filenames):
    nt = min(len(ds.time), 190)
    prognostics = ['QT', 'SLI']
    scm_data = single_column_simulation(
        model, location, interval=(0, nt - 1), prognostics=prognostics)
    merged_pred_data = location.rename({
        'SLI': 'SLIOBS',
        'QT': 'QTOBS',
        'U': 'UOBS',
        'V': 'VOBS',
    }).merge(
        scm_data, join='inner')
    output = model.call_with_xr(merged_pred_data)

    plt.figure()
    scm_data.QT.plot(x='time')
    plt.title("QT")
    plt.savefig(filenames[0])

    plt.figure()
    output.QT.plot(x='time')
    plt.title("FQTNN from NN-prediction")
    output_truth = model.call_with_xr(location)
    plt.savefig(filenames[1])

    plt.figure()
    output_truth.QT.plot(x='time')
    plt.title("FQTNN from Truth")
    plt.xlim([100, 125])
    plt.savefig(filenames[2])

    plt.figure()
    (scm_data.QT * ds.layer_mass / 1000).sum('z').plot()
    (location.QT * ds.layer_mass / 1000).sum('z').plot(label='observed')
    plt.legend()
    plt.savefig(filenames[3])


## Some plotting routines to be called after every epoch
@attr.s
class Plot(object):
    ex = attr.ib()

    def get_filename(self, epoch):
        return self.name.format(epoch)

    def save_figure(self, epoch, location, output):
        fig, ax = plt.subplots()
        self.plot(location, output, ax)
        path = self.get_filename(epoch)
        logging.info(f"Saving to {path}")
        fig.savefig(path)
        self.ex.add_artifact(path)
        plt.close()


class plot_q2(Plot):
    name = 'q2_{}.png'

    def plot(self, location, output, ax):
        return output.QT.plot(x='time', ax=ax)


class plot_scatter_q2_fqt(Plot):
    name = 'scatter_fqt_q2_{}.png'

    def plot(self, location, output, ax):
        x, y, z = [
            x.values.ravel()
            for x in xr.broadcast(location.FQT * 86400, output.QT, output.z)
        ]
        im = ax.scatter(x, y, c=z)
        plt.colorbar(im, ax=ax)


def imbalance_plot(model, dataset, engine):
    from uwnet.thermo import lhf_to_evap
    n = engine.state.epoch
    mass = dataset.layer_mass
    subset = dataset.isel(time=slice(0, None, 20))
    out = model.call_with_xr(subset)
    pme = (subset.Prec - lhf_to_evap(subset.LHF)).mean(['time', 'x'])
    pmenn = -(mass * out.QT).sum('z') / 1000
    pmefqt = ((mass * subset.FQT).sum('z') / 1000 * 86400).mean(['time', 'x'])
    pmenn = pmenn.mean(['time', 'x'])
    plotme = xr.Dataset({
        'truth': pme,
        'nn': pmenn,
        'fqt': pmefqt
    }).to_array(dim='var')

    plt.figure()
    try:
        plotme.plot(hue='var')
    except ValueError:
        df = plotme.squeeze().to_series()
        df.plot(kind='bar')

    plt.savefig(f"{n}-imbalance.png")

    plt.close()



import click
import xarray as xr
from dask.diagnostics import ProgressBar
from uwnet import thermo


def compute_radiation(data):
    lw = data.pop('LWNS')
    sw = data.pop('SWNS')
    data['RADSFC'] = lw - sw

    lw = data.pop('LWNT')
    sw = data.pop('SWNT')
    data['RADTOA'] = lw - sw

    return data


def compute_forcings(data):
    pass


def load_data(paths):

    data = {}
    for info in paths:
        for field in info['fields']:
            try:
                ds = xr.open_dataset(info['path'], chunks={'time': 40})
            except ValueError:
                ds = xr.open_dataset(info['path'])

            data[field] = ds[field]

    # compute layer mass from stat file
    # remove presssur
    rho = data.pop('RHO')[0]
    rhodz = thermo.layer_mass(rho)

    data['layer_mass'] = rhodz

    # compute thermodynamic variables
    TABS = data.pop('TABS')
    QV = data.pop('QV')
    QN = data.get('QN', 0.0)
    QP = data.get('QP', 0.0)

    sl = thermo.liquid_water_temperature(TABS, QN, QP)
    qt = QV + QN

    data['sl'] = sl.assign_attrs(units='K')
    data['qt'] = qt.assign_attrs(units='g/kg')

    data = compute_radiation(data)
    compute_forcings(data)

    # rename staggered dimensions
    data['U'] = (data['U']
                 .rename({'xs': 'x', 'yc': 'y'}))

    data['V'] = (data['V']
                 .rename({'xc': 'x', 'ys': 'y'}))

    objects = [
        val.to_dataset(name=key).assign(x=sl.x, y=sl.y)
        for key, val in data.items()
    ]

    ds = xr.merge(objects, join='inner').sortby('time')
    return ds


@click.command()
@click.argument("config")
@click.argument("out")
def main(config, out):
    import yaml
    paths = yaml.load(open(config))['paths']
    with ProgressBar():
        load_data(paths)\
            .load()\
            .to_zarr(out)


if __name__ == '__main__':
    main()



from typing import Mapping

import numpy as np
import torch


# Basic utilities and calculus
def xarray2numpy(x):
    return np.asarray(x)


def torch2numpy(x):
    if isinstance(x, Mapping):
        return {key: torch2numpy(val) for key, val in x.items()}
    else:
        return x.detach().numpy()


def numpy2torch(x):
    if isinstance(x, Mapping):
        return {key: numpy2torch(val) for key, val in x.items()}
    else:
        return torch.tensor(x, requires_grad=True).float()



"""Routines for opening and analyzing debugging output from SAM/UWNET"""
import glob
import logging
import os
from collections.abc import Mapping

import numpy as np
import pandas as pd

import torch
import xarray as xr


def index_like(x, y):
    if isinstance(x, Mapping):

        keys = set(x) & set(y.data_vars)
        return xr.Dataset({key: index_like(x[key], y[key]) for key in keys})
    else:
        if x.shape[0] == 1:
            x = x[0]
        return xr.DataArray(x, dims=y.dims, coords=y.coords)


def open_debug_state_like_ds(path: str, ds: xr.Dataset) -> xr.Dataset:
    """Open SAM debugging output as xarray object

    Parameters
    ----------
    path
        path to pickle saved by torch
    ds
        dataset to use a template

    Returns
    -------
    state
        dataset with fields from path
    """
    dbg = torch.load(path)
    state = index_like(dbg['args'][0], ds)
    out = index_like(dbg['out'], ds)
    out = out.rename(
        {key: key + 'NN'
         for key in set(state.data_vars) & set(out.data_vars)})
    return xr.merge([state, out])


def open_debug_files_as_numpy(files):
    out = {}

    logger = logging.getLogger(__name__)
    for file in files:
        logger.info(f"Opening {file}")
        args = torch.load(file)['args'][0]
        for key in args:
            out.setdefault(key, []).append(args[key])

    for key in out:
        out[key] = np.stack(out[key])

    return out


def expand_dims(x):
    """Assign values to all data-less dimensions

    Needed by hv.Dataset
    """
    coords = {}
    for k, dim in enumerate(x.dims):
        if dim not in x.coords:
            coords[dim] = np.arange(x.shape[k])
    y = x.assign_coords(**coords)
    y.name = x.name
    return y


def concat_datasets(args, name='mode'):
    """Concatenate datasets with a new named index

    This function is especially useful for comparing two datasets with
    shared variables with holoviews

    Parameters
    ----------
    args : list
        list of (name, dataset) pairs
    name : str
        name of the new dimension

    Returns
    -------
    ds : xr.Dataset
        concatenated data
    """

    names, vals = zip(*args)

    # get list of vars
    vars = set(vals[0].data_vars)
    for val in vals:
        vars = vars & set(val.data_vars)
    vars = list(vars)

    vals = [val[vars] for val in vals]

    return xr.concat(vals, dim=pd.Index(names, name=name))


def open_debug_and_training_data(t, ids, training_data_path):
    """Open an concatenate the debugging and training data"""

    debug_files = {
        tag: glob.glob(os.path.join(path, '*.pkl'))
        for tag, path in ids.items()
    }

    # open training
    training_ds = xr.open_dataset(training_data_path)
    train_ds_init_time = training_ds.isel(time=0)

    args = [('Train', train_ds_init_time)]
    for tag in debug_files:
        dbg = open_debug_state_like_ds(debug_files[tag][t], train_ds_init_time)
        args.append((tag, dbg))
    return concat_datasets(args, name='tag')



import numpy as np
import dask.array as da
import xarray as xr

from .datasets import tiltwave
from .xcalc import centderiv, centspacing, dask_centdiff

PLOT = False


def test_dask_centderiv():
    x, y = np.mgrid[0:2 * np.pi:10j, 0:2 * np.pi:10j]

    z = da.from_array(x, chunks=(10, 10))
    dz = dask_centdiff(z, axis=0, boundary='periodic')

    dx = np.roll(x, -1, 0) - np.roll(x, 1, 0)
    np.testing.assert_allclose(dx, dz.compute())


def test_centdiff():

    A = tiltwave().chunk()

    B = centderiv(A, dim='z')
    B = A.centderiv(dim='z')

    if PLOT:
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        fig, (a, b) = plt.subplots(1, 2)
        B.plot(ax=a)

        A.plot(ax=b)
        plt.show()


def test_centspacing():
    x = np.arange(5)

    xd = xr.DataArray(x, coords=(('x', x), ))

    dx = centspacing(xd).values

    np.testing.assert_allclose(dx, np.ones_like(dx) * 2)



from uwnet.constraints import (apply_linear_constraint, fix_negative_moisture,
                               expected_moisture, enforce_expected_integral,
                               expected_temperature)
import numpy as np
import torch

import pytest


def test_apply_linear_constraint():
    def lin(x):
        """sum(x) >= 0"""
        return x.sum(-1, keepdim=True)

    x = torch.ones(1, 10)

    # test equality constraint
    y = apply_linear_constraint(lin, 0, x).data.numpy()
    np.testing.assert_allclose(y, 0.0)

    # test inequality constraint
    y = apply_linear_constraint(lin, 0, x, inequality=True)
    np.testing.assert_almost_equal(y.data.numpy(), x.data.numpy())

    # test inequality constraint
    y = apply_linear_constraint(lin, 0, -x, inequality=True)
    np.testing.assert_allclose(y.data.numpy(), 0.0)

    # test shape
    assert y.size() == x.size()


def test_fix_negative_moisture():
    q = torch.arange(-4, 20).float() + .1
    layer_mass = torch.rand(len(q)) + .1

    q_new = fix_negative_moisture(q, layer_mass)

    # No negative humidiy
    assert (q_new >= 0).all()

    # water is conserved
    q1 = (q_new * layer_mass).sum(-1)
    q0 = (q * layer_mass).sum(-1)
    np.testing.assert_allclose(q1.numpy(), q0.numpy(), rtol=1e-5)


def test_fix_expected_moisture():
    n = 11
    q = torch.rand(n)
    dt = 86400

    layer_mass = torch.rand(n) + .5

    pw0, pw = expected_moisture(q, 0, 0, 0, 0, layer_mass)
    actual = (layer_mass * q).sum()
    np.testing.assert_allclose(pw.item(), actual.item())

    latent_heat = 2.51e6
    evap = .0005  # kg/m^2/s
    lhf = evap * latent_heat
    pw0, pw = expected_moisture(q, 0, 0, lhf, dt, layer_mass)
    np.testing.assert_allclose((pw - pw0).item() / 1000, evap * dt)


def test_enforce_expected_integral():
    expected = 2.0

    x = torch.rand(100)
    w = torch.rand(100) + .5
    x = enforce_expected_integral(x, expected, w)
    assert (x * w).sum().item() == pytest.approx(expected)


def test_expected_temperature():

    h = 86400
    temp = torch.tensor(300.0)
    mass = torch.tensor(1.0)

    delta_temp = torch.tensor(1.0/86400)

    next_temp = temp + delta_temp * h
    _, ans = expected_temperature(
        temp,
        delta_temp,
        prec=0,
        shf=0,
        radtoa=0,
        radsfc=0,
        layer_mass=mass,
        h=h)
    assert pytest.approx(ans.item()) == next_temp.item()

    # SHF
    shf = 100  # W/m2
    next_temp = temp + shf / 1004 * h
    _, ans = expected_temperature(
        temp, 0, prec=0, shf=shf, radtoa=0, radsfc=0, layer_mass=mass, h=h)
    assert pytest.approx(ans.item()) == next_temp.item()

    # RADTOA
    _, ans = expected_temperature(
        temp, 0, prec=0, shf=0, radtoa=-shf, radsfc=0, layer_mass=mass, h=h)
    assert pytest.approx(ans.item()) == next_temp.item()

    # RADSFC
    _, ans = expected_temperature(
        temp, 0, prec=0, shf=0, radtoa=0, radsfc=shf, layer_mass=mass, h=h)
    assert pytest.approx(ans.item()) == next_temp.item()



from dataclasses import dataclass

import logging
import numpy as np
import os
import xarray as xr

from src.data import assign_apparent_sources

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('dataset_to_skformat_loader')

@dataclass
class XarrayLoaderSklearn:
    """
    Yield X,y arrays for use in models with sklearn_generic style data formatting
    """
    dataset: xr.Dataset
    batch_size: int
    variables: list = None
    dims: tuple = ('sample', 'time', 'z', 'x', 'y')
    existing_data_file: str = None
    write_sample_to_file: bool = False
    output_data_matrices_path: str = None

    @property
    def batches(self):
        n = self.num_samples
        batch_size = self.batch_size
        indices = list(range(0, n, batch_size)) + [n]
        batches = [slice(i_start, i_end)
                   for i_start, i_end in zip(indices[:-1], indices[1:])]
        return batches

    @property
    def num_samples(self):
        if 'sample' in self.dataset.dims:
            return len(self.dataset.sample)
        else:
            return len(self.dataset.x)

    def __iter__(self):
        """
        :param existing_data_file: if provided, will try to load existing .npy saved data
        :param write_to_file_path: if specified, will write the arrays to writ to
        :return:
        """
        for batch in self.batches:
            logger.info( (f'Load Xarray with n samples = {self.num_samples} ') )

            if self.existing_data_file and os.path.exists(self.existing_data_file):
                logger.info(f'Read existing data from {self.existing_data_file}')
                batch_data = np.load(self.existing_data_file)
                num_target_columns = int((np.shape(batch_data)[1] - 2) / 2)
                x_array = batch_data[:, :num_target_columns]
                y_array = batch_data[:, -num_target_columns:]

            else:
                if self.variables is None:
                    variables = list(self.dataset.data_vars)
                else:
                    variables = self.variables

                subset = self.dataset[variables].isel(x=batch)

                # compute apparent sources Q1 and Q2 for y inputs
                subset = assign_apparent_sources(subset)
                x_array = self._stack_dims_and_concat_feats(
                            subset,
                            ['SLI', 'QT', 'SOLIN', 'SST'])
                y_array = self._stack_dims_and_concat_feats(
                            subset,
                            ['Q1', 'Q2'])
                batch_data = np.hstack([x_array, y_array])

                # rm rows with NaN elements
                batch_data = batch_data[~np.isnan(batch_data).any(axis=1)]
                num_target_columns = int((np.shape(batch_data)[1] - 2) / 2)
                x_array = batch_data[:, :-num_target_columns]
                y_array = batch_data[:, -num_target_columns:]

            if self.write_sample_to_file and self.output_data_matrices_path:
                np.save(self.output_data_matrices_path, batch_data)
                logger.info(f'Wrote batch to {self.output_data_matrices_path}')

            yield x_array, y_array

    def _stack_dims_and_concat_feats(
            self,
            ds,
            variables,
            sample_dims=('x', 'y', 'time'),
            feature_dims=('z')
    ):
        """Convert certain variables of a data frame into 2D numpy arrays"""

        # convert tuple args to lists
        sample_dims = list(sample_dims)
        feature_dims = list(feature_dims)

        flat_arrays = []
        for name in variables:
            da = ds[name]

            # for two-d variables insert a singleton "z" dimension
            if 'z' not in da.dims:
                da = da.expand_dims('z')
            stacked_da = da.stack(samples=sample_dims, features=feature_dims)
            # make sure the rows are samples and columns are features
            transposed_da = stacked_da.transpose('samples', 'features')
            flat_arrays.append(transposed_da.values)

        # concatenate along the final dimension
        return np.concatenate(flat_arrays, axis=1)




from src.data import open_ngaqua
import xarray as xr
from scipy.signal import welch

from xgcm import Grid

import numpy as np
import matplotlib.pyplot as plt

fft = np.fft.fft


def interpolated_velocities(u, v):
    """Interpolate the staggered horizontal velocities to the cell centers

    Use's Ryan Abernathy's xgcm package.

    Parameters
    ----------
    u, v : xr.DataArray
        staggered velocities.

    Returns
    -------
    uint, vint : xr.DataArray
        interpolated velocites on (xc, yc) grid
    """
    # setup grid
    xl = u.x.values
    xc = xl + (xl[1] - xl[0]) / 2

    yl = u.y.values
    yc = yl + (yl[1] - yl[0]) / 2

    u = u.rename({'x': 'xl', 'y': 'yc'}).assign_coords(xl=xl, yc=yc)
    v = v.rename({'x': 'xc', 'y': 'yl'}).assign_coords(yl=yl, xc=xc)

    # combine the data
    ds = xr.Dataset({'u': u, 'v': v})

    # create grid object
    coords = {
        'x': {
            'center': 'xc',
            'left': 'xl'
        },
        'y': {
            'center': 'yc',
            'left': 'yl'
        }
    }
    grid = Grid(ds, periodic=['x'], coords=coords)

    # use grid to interpolate
    uint = grid.interp(ds.u, axis='x')
    vint = grid.interp(ds.v, axis='y', boundary='extend')

    return uint, vint


def horizontal_power_spectrum(u, dim='xc', avg=('yc', 'time')):
    dim_vals = u[dim].values

    d = dim_vals[1] - dim_vals[0]
    fu = fft(u, axis=u.get_axis_num(dim)).mean(u.get_axis_num(avg))
    pw = np.abs(fu)**2

    k = np.fft.fftfreq(n=len(fu), d=d)
    k, pw = k[k >= 0], pw[k >= 0]

    return xr.DataArray(pw, dims=['freq_x'], coords=[k])


def horizontal_welch(u, dim='xc', avg=('yc', 'time')):
    dim_vals = u[dim].values
    d = dim_vals[1] - dim_vals[0]
    k, pw = welch(u, fs=1 / d, axis=u.get_axis_num(dim), scaling='spectrum')
    pw = pw.mean(u.get_axis_num(avg))
    k, pw = k[k >= 0], pw[k >= 0]
    return xr.DataArray(pw, dims=['freq_x'], coords=[k])


def kinetic_energy_spectrum(uint, vint):
    """Compute Kinetic Energy Spectra"""
    pw_u = horizontal_welch(uint)
    pw_v = horizontal_welch(vint)
    return (pw_u + pw_v) / 2.0


def kinetic_energy_spectrum_staggered(*args):
    vels = interpolated_velocities(*args)
    return kinetic_energy_spectrum(*vels)


def autocorr(x: xr.DataArray,
             dim: str,
             base_state_dims=(),
             avg_dims=(),
             n=None) -> xr.DataArray:
    """Auto-correlation of a time series

    Parameters
    ----------
    x : xr.DataArray
        Array to compute auto-correlation of
    dim : str
        Dimensions to copmute auto-correlation over
    base_state_dims : Sequence[str]
        List of dimensions

    Returns
    -------
    autocorrelation
    """

    base_state_dims = (dim, ) + tuple(base_state_dims)
    avg_dims = base_state_dims + tuple(avg_dims)

    if n is None:
        n = len(x[dim])
    lags = np.arange(1, n // 2)

    time = x[dim]
    # assume constant spacing
    dt = float(time[1] - time[0])

    x = x - x.mean(base_state_dims)

    denom = (x * x).sum(avg_dims)

    corrs = []
    for lag in lags:
        shift = {dim: lag}
        xs = x.shift(**shift)
        corr = (xs * x).dropna(dim).sum(avg_dims) / denom
        corrs.append(corr)

    lags = xr.DataArray(
        lags * dt, name='lag', attrs={'units': 'day'}, dims=['lag'])

    return xr.concat(corrs, dim=lags)



import numpy as np
import pytest
import xarray as xr

from uwnet.ml_models.nn.datasets_handler import XRTimeSeries, get_timestep, XarrayBatchLoader


def get_obj():

    dims_3d = ['time', 'z', 'y', 'x']
    dims_2d = ['time', 'y', 'x']

    data_3d = np.ones((4, 4, 5, 2))
    data_2d = np.ones((4, 5, 2))

    return xr.Dataset({
        'a': (dims_3d, data_3d),
        'b': (dims_2d, data_2d)
    }), data_3d.shape


@pytest.mark.parametrize('time_length', [1, 2, 4])
def test_XRTimeSeries(time_length):
    ds, (t, z, y, x) = get_obj()
    o = XRTimeSeries(ds, time_length=time_length)
    assert len(o) == t * y * x // time_length


def test_XRTimeSeries_shape():
    ds, (t, z, y, x) = get_obj()

    time_length = 2
    o = XRTimeSeries(ds, time_length=time_length)
    assert o[0]['a'].shape == (time_length, z, 1, 1)
    assert o[0]['b'].shape == (time_length, 1, 1, 1)

    # get last time point
    o[-1]


@pytest.mark.parametrize('dt,units,dt_seconds', [
    (.125, 'd', 10800.0),
    (10, 's', 10.0),
])
def test_get_timestep(dt, units, dt_seconds):
    ds, _ = get_obj()
    ds = ds.assign_coords(time=np.arange(len(ds['time'])) * dt)
    ds['time'].attrs['units'] = units

    dt = get_timestep(ds)
    assert dt == dt_seconds

    # non-uniformly sampled data
    time = np.arange(len(ds['time']))**2
    ds = ds.assign_coords(time=time)
    with pytest.raises(ValueError):
        get_timestep(ds)


@pytest.mark.xfail()
def test_XRTimeSeries_torch_constants():
    ds = get_obj()

    # dim_1 is the time dimension here
    dataset = XRTimeSeries(ds, [['dim_1'], ['dim_0', 'dim_2'], ['dim_3']])
    assert 'layer_mass' in dataset.torch_constants()

    # constants should not be in batch
    assert 'layer_mass' not in dataset[0]


def _init_XarrayBatchLoader(num_samples, batch_size, num_time):

    shape = (num_samples, num_time, 2)
    s, t, z = shape
    ds = xr.Dataset({
        'a': (['sample', 'time', 'z'], np.ones(shape))
    })
    return XarrayBatchLoader(ds, batch_size=batch_size)


@pytest.mark.parametrize('num_samples, batch_size',[
    (10, 1),
    (10, 3),
])
def test_XarrayBatchLoader_batches(num_samples, batch_size):

    num_time = 12
    loader = _init_XarrayBatchLoader(num_samples, batch_size, num_time)
    batches = loader.batches
    last_batch_index = batches[-1].stop
    assert last_batch_index == num_samples


def check_batch(batch, num_time):
    for key, arr in batch.items():
        assert arr.ndim == 5
        assert arr.shape[1] == num_time


def test_XarrayBatchLoader_iter():
    num_samples, batch_size, num_time = 10, 3, 4
    loader = _init_XarrayBatchLoader(num_samples, batch_size, num_time)

    # see if the loader works
    for batch in loader:
        check_batch(batch, num_time)
